import configparser
import logging
import os
import re
import shutil
import time

import pytest
from saltfactories.utils.functional import Loaders

import salt.utils.path
import salt.utils.pkg
import salt.utils.platform

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.timeout_unless_on_windows(240),
]


@pytest.fixture
def ctx():
    return {}


@pytest.fixture
def _preserve_rhel_yum_conf(tmp_path):

    # save off current yum.conf
    cfg_file = "/etc/yum.conf"
    if not os.path.exists(cfg_file):
        pytest.skip("Only runs on RedHat.")

    tmp_file = tmp_path / "yum.conf"
    shutil.copy2(cfg_file, tmp_file)
    try:
        yield
    finally:
        # restore saved yum.conf
        shutil.copy2(tmp_file, cfg_file)


@pytest.fixture
def _refresh_db(ctx, grains, modules):
    if "refresh" not in ctx:
        modules.pkg.refresh_db()
        ctx["refresh"] = True

    # If this is Arch Linux, check if pacman is in use by another process
    if grains["os_family"] == "Arch":
        for _ in range(12):
            if not os.path.isfile("/var/lib/pacman/db.lck"):
                break
            else:
                time.sleep(5)
        else:
            raise Exception("Package database locked after 60 seconds, bailing out")


@pytest.fixture(autouse=True)
def test_pkg(grains):
    _pkg = "figlet"
    if salt.utils.platform.is_windows():
        _pkg = "putty"
    elif grains["os_family"] == "RedHat":
        if grains["os"] == "VMware Photon OS":
            _pkg = "snoopy"
        elif grains["osfinger"] == "Amazon Linux-2023":
            return "dnf-utils"
        else:
            _pkg = "units"
    elif grains["os_family"] == "Debian":
        _pkg = "ifenslave"
    return _pkg


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.requires_salt_modules("pkg.list_pkgs")
@pytest.mark.slow_test
def test_list(modules):
    """
    verify that packages are installed
    """
    ret = modules.pkg.list_pkgs()
    assert len(ret.keys()) != 0


@pytest.mark.requires_salt_modules("pkg.version_cmp")
@pytest.mark.slow_test
def test_version_cmp(grains, modules):
    """
    test package version comparison on supported platforms
    """
    if grains["os_family"] == "Debian":
        lt = ["0.2.4-0ubuntu1", "0.2.4.1-0ubuntu1"]
        eq = ["0.2.4-0ubuntu1", "0.2.4-0ubuntu1"]
        gt = ["0.2.4.1-0ubuntu1", "0.2.4-0ubuntu1"]
    elif grains["os_family"] == "Suse":
        lt = ["2.3.0-1", "2.3.1-15.1"]
        eq = ["2.3.1-15.1", "2.3.1-15.1"]
        gt = ["2.3.2-15.1", "2.3.1-15.1"]
    else:
        lt = ["2.3.0", "2.3.1"]
        eq = ["2.3.1", "2.3.1"]
        gt = ["2.3.2", "2.3.1"]

    assert modules.pkg.version_cmp(*lt) == -1
    assert modules.pkg.version_cmp(*eq) == 0
    assert modules.pkg.version_cmp(*gt) == 1


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.destructive_test
@pytest.mark.requires_salt_modules("pkg.mod_repo", "pkg.del_repo", "pkg.get_repo")
@pytest.mark.slow_test
@pytest.mark.requires_network
def test_mod_del_repo(grains, modules):
    """
    test modifying and deleting a software repository
    """
    repo = None

    try:
        # ppa:otto-kesselgulasch/gimp-edge has no Ubuntu 22.04 repo
        if grains["os"] == "Ubuntu" and grains["osmajorrelease"] < 22:
            repo = "ppa:otto-kesselgulasch/gimp-edge"
            uri = "http://ppa.launchpad.net/otto-kesselgulasch/gimp-edge/ubuntu"
            ret = modules.pkg.mod_repo(repo, "comps=main")
            assert ret != []
            ret = modules.pkg.get_repo(repo)

            assert isinstance(ret, dict) is True
            assert ret["uri"] == uri
        elif grains["os_family"] == "RedHat":
            repo = "saltstack"
            name = "SaltStack repo for RHEL/CentOS {}".format(grains["osmajorrelease"])
            baseurl = "https://packages.broadcom.com/artifactory/saltproject-rpm/"
            gpgkey = "https://packages.broadcom.com/artifactory/api/security/keypair/SaltProjectKey/public"
            gpgcheck = 1
            enabled = 1
            ret = modules.pkg.mod_repo(
                repo,
                name=name,
                baseurl=baseurl,
                gpgkey=gpgkey,
                gpgcheck=gpgcheck,
                enabled=enabled,
            )
            # return data from pkg.mod_repo contains the file modified at
            # the top level, so use next(iter(ret)) to get that key
            assert ret != {}
            repo_info = ret[next(iter(ret))]
            assert repo in repo_info
            assert repo_info[repo]["baseurl"] == baseurl
            ret = modules.pkg.get_repo(repo)
            assert ret["baseurl"] == baseurl
    finally:
        if repo is not None:
            modules.pkg.del_repo(repo)


@pytest.mark.slow_test
@pytest.mark.usefixtures("_refresh_db")
def test_mod_del_repo_multiline_values(modules):
    """
    test modifying and deleting a software repository defined with multiline values
    """
    os_grain = modules.grains.item("os")["os"]
    repo = None
    try:
        if os_grain in ["CentOS", "RedHat", "VMware Photon OS"]:
            my_baseurl = (
                "http://my.fake.repo/foo/bar/\n http://my.fake.repo.alt/foo/bar/"
            )
            expected_get_repo_baseurl = (
                "http://my.fake.repo/foo/bar/\nhttp://my.fake.repo.alt/foo/bar/"
            )
            repo = "fakerepo"
            name = "Fake repo for RHEL/CentOS/SUSE"
            baseurl = my_baseurl
            gpgkey = "https://my.fake.repo/foo/bar/MY-GPG-KEY.pub"
            failovermethod = "priority"
            gpgcheck = 1
            enabled = 1
            ret = modules.pkg.mod_repo(
                repo,
                name=name,
                baseurl=baseurl,
                gpgkey=gpgkey,
                gpgcheck=gpgcheck,
                enabled=enabled,
                failovermethod=failovermethod,
            )
            # return data from pkg.mod_repo contains the file modified at
            # the top level, so use next(iter(ret)) to get that key
            assert ret != {}
            repo_info = ret[next(iter(ret))]
            assert repo in repo_info
            assert repo_info[repo]["baseurl"] == my_baseurl
            ret = modules.pkg.get_repo(repo)
            assert ret["baseurl"] == expected_get_repo_baseurl
            modules.pkg.mod_repo(repo)
            ret = modules.pkg.get_repo(repo)
            assert ret["baseurl"] == expected_get_repo_baseurl
    finally:
        if repo is not None:
            modules.pkg.del_repo(repo)


@pytest.mark.requires_salt_modules("pkg.owner")
def test_owner(modules, grains):
    """
    test finding the package owning a file
    """
    binary = "/bin/ls"
    if grains["os"] == "Ubuntu" and grains["osmajorrelease"] >= 24:
        binary = "/usr/bin/ls"
    if grains["os"] == "Debian" and grains["osmajorrelease"] >= 13:
        binary = "/usr/bin/ls"

    ret = modules.pkg.owner(binary)
    assert len(ret) != 0


# Similar to pkg.owner, but for FreeBSD's pkgng
@pytest.mark.skip_on_freebsd(reason="test for new package manager for FreeBSD")
@pytest.mark.requires_salt_modules("pkg.which")
def test_which(modules, grains):
    """
    test finding the package owning a file
    """
    binary = "/bin/ls"
    if grains["os"] == "Ubuntu" and grains["osmajorrelease"] >= 24:
        binary = "/usr/bin/ls"
    ret = modules.pkg.which(binary)
    assert len(ret) != 0


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.destructive_test
@pytest.mark.requires_salt_modules("pkg.version", "pkg.install", "pkg.remove")
@pytest.mark.slow_test
@pytest.mark.requires_network
def test_install_remove(modules, test_pkg):
    """
    successfully install and uninstall a package
    """
    version = modules.pkg.version(test_pkg)

    def test_install():
        install_ret = modules.pkg.install(test_pkg)
        assert test_pkg in install_ret

    def test_remove():
        remove_ret = modules.pkg.remove(test_pkg)
        assert test_pkg in remove_ret

    if version and isinstance(version, dict):
        version = version[test_pkg]

    if version:
        test_remove()
        test_install()
    else:
        test_install()
        test_remove()


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.destructive_test
@pytest.mark.skip_on_photonos(
    reason="package hold/unhold unsupported on Photon OS",
)
@pytest.mark.requires_salt_modules(
    "pkg.hold",
    "pkg.unhold",
    "pkg.install",
    "pkg.version",
    "pkg.remove",
    "pkg.list_pkgs",
)
@pytest.mark.slow_test
@pytest.mark.requires_network
@pytest.mark.requires_salt_states("pkg.installed")
def test_hold_unhold(grains, modules, states, test_pkg):
    """
    test holding and unholding a package
    """
    versionlock_pkg = None
    if grains["os_family"] == "RedHat":
        pkgs = {p for p in modules.pkg.list_repo_pkgs() if "-versionlock" in p}
        if not pkgs:
            pytest.skip("No versionlock package found in repositories")
        for versionlock_pkg in pkgs:
            ret = states.pkg.installed(name=versionlock_pkg, refresh=False)
            # Exit loop if a versionlock package installed correctly
            try:
                assert ret.result is True
                break
            except AssertionError:
                pass
        else:
            pytest.fail(f"Could not install versionlock package from {pkgs}")

    modules.pkg.install(test_pkg)

    try:
        hold_ret = modules.pkg.hold(test_pkg)
        if versionlock_pkg and "-versionlock is not installed" in str(hold_ret):
            pytest.skip(f"{hold_ret}  `{versionlock_pkg}` is installed")
        assert test_pkg in hold_ret
        assert hold_ret[test_pkg]["result"] is True

        unhold_ret = modules.pkg.unhold(test_pkg)
        assert test_pkg in unhold_ret
        assert unhold_ret[test_pkg]["result"] is True
        modules.pkg.remove(test_pkg)
    except salt.exceptions.SaltInvocationError as err:
        if "versionlock is not installed" in err.message:
            pytest.skip("Correct versionlock package is not installed")
    finally:
        if versionlock_pkg:
            ret = states.pkg.removed(name=versionlock_pkg)
            assert ret.result is True


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.destructive_test
@pytest.mark.requires_salt_modules("pkg.refresh_db")
@pytest.mark.slow_test
@pytest.mark.requires_network
def test_refresh_db(grains, minion_opts):
    """
    test refreshing the package database
    """
    rtag = salt.utils.pkg.rtag(minion_opts)
    salt.utils.pkg.write_rtag(minion_opts)
    assert os.path.isfile(rtag) is True

    loader = Loaders(minion_opts)
    ret = loader.modules.pkg.refresh_db()
    if not isinstance(ret, dict):
        pytest.skip(f"Upstream repo did not return coherent results: {ret}")

    if grains["os_family"] == "RedHat":
        assert ret in (True, None)
    elif grains["os_family"] == "Suse":
        if not isinstance(ret, dict):
            pytest.skip("Upstream repo did not return coherent results. Skipping test.")
        assert ret != {}
        for source, state in ret.items():
            assert state in (True, False, None)

    assert os.path.isfile(rtag) is False


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.requires_salt_modules("pkg.info_installed")
@pytest.mark.slow_test
def test_pkg_info(grains, modules, test_pkg):
    """
    Test returning useful information on Ubuntu systems.
    """
    if grains["os_family"] == "Debian":
        ret = modules.pkg.info_installed("bash", "dpkg")
        keys = ret.keys()
        assert "bash" in keys
        assert "dpkg" in keys
    elif grains["os_family"] == "RedHat":
        ret = modules.pkg.info_installed("rpm", "bash")
        keys = ret.keys()
        assert "rpm" in keys
        assert "bash" in keys
    elif grains["os_family"] == "Suse":
        ret = modules.pkg.info_installed("less", "zypper")
        keys = ret.keys()
        assert "less" in keys
        assert "zypper" in keys
    else:
        ret = modules.pkg.info_installed(test_pkg)
        keys = ret.keys()
        assert test_pkg in keys


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.skipif(True, reason="Temporary Skip - Causes centos 8 test to fail")
@pytest.mark.destructive_test
@pytest.mark.requires_salt_modules(
    "pkg.refresh_db",
    "pkg.upgrade",
    "pkg.install",
    "pkg.list_repo_pkgs",
    "pkg.list_upgrades",
)
@pytest.mark.slow_test
@pytest.mark.requires_network
def test_pkg_upgrade_has_pending_upgrades(grains, modules):
    """
    Test running a system upgrade when there are packages that need upgrading
    """
    if grains["os"] == "Arch":
        pytest.skip("Arch moved to Python 3.8 and we're not ready for it yet")

    modules.pkg.upgrade()

    # First make sure that an up-to-date copy of the package db is available
    modules.pkg.refresh_db()

    if grains["os_family"] == "Suse":
        # This test assumes that there are multiple possible versions of a
        # package available. That makes it brittle if you pick just one
        # target, as changes in the available packages will break the test.
        # Therefore, we'll choose from several packages to make sure we get
        # one that is suitable for this test.
        packages = ("hwinfo", "avrdude", "diffoscope", "vim")
        available = modules.pkg.list_repo_pkgs(packages)

        for package in packages:
            try:
                new, old = available[package][:2]
            except (KeyError, ValueError):
                # Package not available, or less than 2 versions
                # available. This is not a suitable target.
                continue
            else:
                target = package
                break
        else:
            # None of the packages have more than one version available, so
            # we need to find new package(s). pkg.list_repo_pkgs can be
            # used to get an overview of the available packages. We should
            # try to find packages with few dependencies and small download
            # sizes, to keep this test from taking longer than necessary.
            pytest.fail("No suitable package found for this test")

        # Make sure we have the 2nd-oldest available version installed
        ret = modules.pkg.install(target, version=old)
        if not isinstance(ret, dict):
            if ret.startswith("ERROR"):
                pytest.skip(f"Could not install older {target} to complete test.")

        # Run a system upgrade, which should catch the fact that the
        # targeted package needs upgrading, and upgrade it.
        ret = modules.pkg.upgrade()

        # The changes dictionary should not be empty.
        if "changes" in ret:
            assert target in ret["changes"]
        else:
            assert target in ret
    else:
        ret = modules.pkg.list_upgrades()
        if ret == "" or ret == {}:
            pytest.skip(
                "No updates available for this machine.  Skipping pkg.upgrade test."
            )
        else:
            args = []
            if grains["os_family"] == "Debian":
                args = ["dist_upgrade=True"]
            ret = modules.pkg.upgrade(args)
            assert ret != {}


@pytest.mark.usefixtures("_refresh_db")
@pytest.mark.destructive_test
@pytest.mark.skip_on_darwin(
    reason="The jenkins user is equivalent to root on mac, causing the test to be unrunnable"
)
@pytest.mark.requires_salt_modules("pkg.remove", "pkg.latest_version")
@pytest.mark.slow_test
@pytest.mark.requires_salt_states("pkg.removed")
def test_pkg_latest_version(grains, modules, states, test_pkg):
    """
    Check that pkg.latest_version returns the latest version of the uninstalled package.
    The package is not installed. Only the package version is checked.
    """
    states.pkg.removed(test_pkg)

    cmd_pkg = []
    if grains["os_family"] == "RedHat":
        cmd_pkg = modules.cmd.run(f"yum list {test_pkg}")
    elif salt.utils.platform.is_windows():
        cmd_pkg = modules.pkg.list_available(test_pkg)
    elif grains["os_family"] == "Debian":
        cmd_pkg = modules.cmd.run(f"apt list {test_pkg}")
    elif grains["os_family"] == "Arch":
        cmd_pkg = modules.cmd.run(f"pacman -Si {test_pkg}")
    elif grains["os_family"] == "FreeBSD":
        cmd_pkg = modules.cmd.run(f"pkg search -S name -qQ version -e {test_pkg}")
    elif grains["os_family"] == "Suse":
        cmd_pkg = modules.cmd.run(f"zypper info {test_pkg}")
    elif grains["os_family"] == "MacOS":
        brew_bin = salt.utils.path.which("brew")
        mac_user = modules.file.get_user(brew_bin)
        if mac_user == "root":
            pytest.skip(
                "brew cannot run as root, try a user in {}".format(
                    os.listdir("/Users/")
                )
            )
        cmd_pkg = modules.cmd.run(f"brew info {test_pkg}", run_as=mac_user)
    else:
        pytest.skip("TODO: test not configured for {}".format(grains["os_family"]))
    pkg_latest = modules.pkg.latest_version(test_pkg)
    assert pkg_latest in cmd_pkg


@pytest.mark.usefixtures("_preserve_rhel_yum_conf")
@pytest.mark.destructive_test
@pytest.mark.requires_salt_modules("pkg.list_repos")
@pytest.mark.slow_test
def test_list_repos_duplicate_entries(grains, modules):
    """
    test duplicate entries in /etc/yum.conf

    This is a destructive test as it installs and then removes a package
    """
    if grains["os_family"] != "RedHat":
        pytest.skip("Only runs on RedHat.")

    if grains["os"] == "Amazon":
        pytest.skip("Only runs on RedHat, Amazon /etc/yum.conf differs.")

    # write valid config with duplicates entries
    cfg_file = "/etc/yum.conf"
    with salt.utils.files.fpopen(cfg_file, "w", mode=0o644) as fp_:
        fp_.write("[main]\n")
        fp_.write("gpgcheck=1\n")
        fp_.write("installonly_limit=3\n")
        fp_.write("clean_requirements_on_remove=True\n")
        fp_.write("best=True\n")
        fp_.write("skip_if_unavailable=False\n")
        fp_.write("http_caching=True\n")
        fp_.write("http_caching=True\n")

    ret = modules.pkg.list_repos(strict_config=False)
    assert ret != []
    assert isinstance(ret, dict) is True

    # test explicitly strict_config
    expected = "While reading from '/etc/yum.conf' [line  8]: option 'http_caching' in section 'main' already exists"
    with pytest.raises(configparser.DuplicateOptionError) as exc_info:
        modules.pkg.list_repos(strict_config=True)
    assert str(exc_info.value) == expected

    # test implicitly strict_config
    with pytest.raises(configparser.DuplicateOptionError) as exc_info:
        modules.pkg.list_repos()
    assert str(exc_info.value) == expected


@pytest.mark.destructive_test
@pytest.mark.slow_test
def test_pkg_install_port(grains, modules):
    """
    test install package with a port in the url
    """
    pkgs = modules.pkg.list_pkgs()
    nano = pkgs.get("nano")
    if nano:
        modules.pkg.remove("nano")

    if grains["os_family"] == "Debian":
        url = modules.cmd.run("apt download --print-uris nano").split()[-4]
        if url.startswith("'mirror+file"):
            url = "http://ftp.debian.org/debian/pool/" + url.split("pool")[1].rstrip(
                "'"
            )
        try:
            ret = modules.pkg.install(sources=f'[{{"nano":{url}}}]')
            version = re.compile(r"\d\.\d")
            assert version.search(url).group(0) in ret["nano"]["new"]
        finally:
            modules.pkg.remove("nano")
            if nano:
                # If nano existed on the machine before the test ran
                # re-install that version
                modules.pkg.install(f"nano={nano}")



import copy
import io
import os
import shutil
import sys
import tempfile
import textwrap

import pytest

import salt.config
import salt.loader
import salt.utils.files
from salt.state import HighState
from salt.utils.pydsl import PyDslError
from tests.support.helpers import with_tempdir
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

REQUISITES = ["require", "require_in", "use", "use_in", "watch", "watch_in"]


class CommonTestCaseBoilerplate(TestCase):
    def setUp(self):
        self.root_dir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.addCleanup(shutil.rmtree, self.root_dir, ignore_errors=True)
        self.state_tree_dir = os.path.join(self.root_dir, "state_tree")
        self.cache_dir = os.path.join(self.root_dir, "cachedir")
        if not os.path.isdir(self.root_dir):
            os.makedirs(self.root_dir)

        if not os.path.isdir(self.state_tree_dir):
            os.makedirs(self.state_tree_dir)

        if not os.path.isdir(self.cache_dir):
            os.makedirs(self.cache_dir)
        self.config = salt.config.minion_config(None)
        self.config["root_dir"] = self.root_dir
        self.config["state_events"] = False
        self.config["id"] = "match"
        self.config["file_client"] = "local"
        self.config["file_roots"] = dict(base=[self.state_tree_dir])
        self.config["cachedir"] = self.cache_dir
        self.config["test"] = False
        self.config["grains"] = salt.loader.grains(self.config)
        self.HIGHSTATE = HighState(self.config)
        self.HIGHSTATE.push_active()

    def tearDown(self):
        try:
            self.HIGHSTATE.pop_active()
        except IndexError:
            pass
        del self.config
        del self.HIGHSTATE

    def state_highstate(self, state, dirpath):
        opts = copy.copy(self.config)
        opts["file_roots"] = dict(base=[dirpath])
        HIGHSTATE = HighState(opts)
        HIGHSTATE.push_active()
        try:
            high, errors = HIGHSTATE.render_highstate(state)
            if errors:
                import pprint

                pprint.pprint("\n".join(errors))
                pprint.pprint(high)

            out = HIGHSTATE.state.call_high(high)
            # pprint.pprint(out)
        finally:
            HIGHSTATE.pop_active()


class PyDSLRendererTestCase(CommonTestCaseBoilerplate):
    """
    WARNING: If tests in here are flaky, they may need
    to be moved to their own class. Sharing HighState, especially
    through setUp/tearDown can create dangerous race conditions!
    """

    def render_sls(self, content, sls="", saltenv="base", **kws):
        if "env" in kws:
            # "env" is not supported; Use "saltenv".
            kws.pop("env")

        return self.HIGHSTATE.state.rend["pydsl"](
            io.StringIO(content), saltenv=saltenv, sls=sls, **kws
        )

    @pytest.mark.slow_test
    def test_state_declarations(self):
        result = self.render_sls(
            textwrap.dedent(
                """
            state('A').cmd.run('ls -la', cwd='/var/tmp')
            state().file.managed('myfile.txt', source='salt://path/to/file')
            state('X').cmd('run', 'echo hello world', cwd='/')

            a_cmd = state('A').cmd
            a_cmd.run(shell='/bin/bash')
            state('A').service.running(name='apache')
        """
            )
        )
        self.assertTrue("A" in result and "X" in result)
        A_cmd = result["A"]["cmd"]
        self.assertEqual(A_cmd[0], "run")
        self.assertEqual(A_cmd[1]["name"], "ls -la")
        self.assertEqual(A_cmd[2]["cwd"], "/var/tmp")
        self.assertEqual(A_cmd[3]["shell"], "/bin/bash")

        A_service = result["A"]["service"]
        self.assertEqual(A_service[0], "running")
        self.assertEqual(A_service[1]["name"], "apache")

        X_cmd = result["X"]["cmd"]
        self.assertEqual(X_cmd[0], "run")
        self.assertEqual(X_cmd[1]["name"], "echo hello world")
        self.assertEqual(X_cmd[2]["cwd"], "/")

        del result["A"]
        del result["X"]
        self.assertEqual(len(result), 2)
        # 2 rather than 1 because pydsl adds an extra no-op state
        # declaration.

        s_iter = iter(result.values())
        try:
            s = next(s_iter)["file"]
        except KeyError:
            s = next(s_iter)["file"]
        self.assertEqual(s[0], "managed")
        self.assertEqual(s[1]["name"], "myfile.txt")
        self.assertEqual(s[2]["source"], "salt://path/to/file")

    @pytest.mark.slow_test
    def test_requisite_declarations(self):
        result = self.render_sls(
            textwrap.dedent(
                """
            state('X').cmd.run('echo hello')
            state('A').cmd.run('mkdir tmp', cwd='/var')
            state('B').cmd.run('ls -la', cwd='/var/tmp') \
                        .require(state('X').cmd) \
                        .require(cmd='A') \
                        .watch(service='G')
            state('G').service.running(name='collectd')
            state('G').service.watch_in(state('A').cmd)

            state('H').cmd.require_in(cmd='echo hello')
            state('H').cmd.run('echo world')
        """
            )
        )
        self.assertEqual(len(result), 6)
        self.assertTrue(set("X A B G H".split()).issubset(set(result.keys())))
        b = result["B"]["cmd"]
        self.assertEqual(b[0], "run")
        self.assertEqual(b[1]["name"], "ls -la")
        self.assertEqual(b[2]["cwd"], "/var/tmp")
        self.assertEqual(b[3]["require"][0]["cmd"], "X")
        self.assertEqual(b[4]["require"][0]["cmd"], "A")
        self.assertEqual(b[5]["watch"][0]["service"], "G")
        self.assertEqual(result["G"]["service"][2]["watch_in"][0]["cmd"], "A")
        self.assertEqual(result["H"]["cmd"][1]["require_in"][0]["cmd"], "echo hello")

    @pytest.mark.slow_test
    def test_include_extend(self):
        result = self.render_sls(
            textwrap.dedent(
                """
            include(
                'some.sls.file',
                'another.sls.file',
                'more.sls.file',
                delayed=True
            )
            A = state('A').cmd.run('echo hoho', cwd='/')
            state('B').cmd.run('echo hehe', cwd='/')
            extend(
                A,
                state('X').cmd.run(cwd='/a/b/c'),
                state('Y').file('managed', name='a_file.txt'),
                state('Z').service.watch(file='A')
            )
        """
            )
        )
        self.assertEqual(len(result), 4)
        self.assertEqual(
            result["include"],
            [
                {"base": sls}
                for sls in ("some.sls.file", "another.sls.file", "more.sls.file")
            ],
        )
        extend = result["extend"]
        self.assertEqual(extend["X"]["cmd"][0], "run")
        self.assertEqual(extend["X"]["cmd"][1]["cwd"], "/a/b/c")
        self.assertEqual(extend["Y"]["file"][0], "managed")
        self.assertEqual(extend["Y"]["file"][1]["name"], "a_file.txt")
        self.assertEqual(len(extend["Z"]["service"]), 1)
        self.assertEqual(extend["Z"]["service"][0]["watch"][0]["file"], "A")

        self.assertEqual(result["B"]["cmd"][0], "run")
        self.assertTrue("A" not in result)
        self.assertEqual(extend["A"]["cmd"][0], "run")

    @pytest.mark.slow_test
    def test_cmd_call(self):
        result = self.HIGHSTATE.state.call_template_str(
            textwrap.dedent(
                """\
            #!pydsl
            state('A').cmd.run('echo this is state A', cwd='/')

            some_var = 12345
            def do_something(a, b, *args, **kws):
                return dict(result=True, changes={'a': a, 'b': b, 'args': args, 'kws': kws, 'some_var': some_var})

            state('C').cmd.call(do_something, 1, 2, 3, x=1, y=2) \
                          .require(state('A').cmd)

            state('G').cmd.wait('echo this is state G', cwd='/') \
                          .watch(state('C').cmd)
        """
            )
        )
        ret = next(result[k] for k in result.keys() if "do_something" in k)
        changes = ret["changes"]
        self.assertEqual(
            changes, dict(a=1, b=2, args=(3,), kws=dict(x=1, y=2), some_var=12345)
        )

        ret = next(result[k] for k in result.keys() if "-G_" in k)
        self.assertEqual(ret["changes"]["stdout"], "this is state G")

    @pytest.mark.slow_test
    def test_multiple_state_func_in_state_mod(self):
        with self.assertRaisesRegex(PyDslError, "Multiple state functions"):
            self.render_sls(
                textwrap.dedent(
                    """
                state('A').cmd.run('echo hoho')
                state('A').cmd.wait('echo hehe')
            """
                )
            )

    @pytest.mark.slow_test
    def test_no_state_func_in_state_mod(self):
        with self.assertRaisesRegex(PyDslError, "No state function specified"):
            self.render_sls(
                textwrap.dedent(
                    """
                state('B').cmd.require(cmd='hoho')
            """
                )
            )

    @pytest.mark.slow_test
    def test_load_highstate(self):
        result = self.render_sls(
            textwrap.dedent(
                '''
            import salt.utils.yaml
            __pydsl__.load_highstate(salt.utils.yaml.safe_load("""
            A:
              cmd.run:
                - name: echo hello
                - cwd: /
            B:
              pkg:
                - installed
              service:
                - running
                - require:
                  - pkg: B
                - watch:
                  - cmd: A
            """))

            state('A').cmd.run(name='echo hello world')
            '''
            )
        )
        self.assertEqual(len(result), 3)
        self.assertEqual(result["A"]["cmd"][0], "run")
        self.assertIn({"name": "echo hello"}, result["A"]["cmd"])
        self.assertIn({"cwd": "/"}, result["A"]["cmd"])
        self.assertIn({"name": "echo hello world"}, result["A"]["cmd"])
        self.assertEqual(len(result["A"]["cmd"]), 4)

        self.assertEqual(len(result["B"]["pkg"]), 1)
        self.assertEqual(result["B"]["pkg"][0], "installed")

        self.assertEqual(result["B"]["service"][0], "running")
        self.assertIn({"require": [{"pkg": "B"}]}, result["B"]["service"])
        self.assertIn({"watch": [{"cmd": "A"}]}, result["B"]["service"])
        self.assertEqual(len(result["B"]["service"]), 3)

    @pytest.mark.slow_test
    def test_ordered_states(self):
        result = self.render_sls(
            textwrap.dedent(
                """
            __pydsl__.set(ordered=True)
            A = state('A')
            state('B').cmd.run('echo bbbb')
            A.cmd.run('echo aaa')
            state('B').cmd.run(cwd='/')
            state('C').cmd.run('echo ccc')
            state('B').file.managed(source='/a/b/c')
            """
            )
        )
        self.assertEqual(len(result["B"]["cmd"]), 3)
        self.assertEqual(result["A"]["cmd"][1]["require"][0]["cmd"], "B")
        self.assertEqual(result["C"]["cmd"][1]["require"][0]["cmd"], "A")
        self.assertEqual(result["B"]["file"][1]["require"][0]["cmd"], "C")

    @with_tempdir()
    @pytest.mark.slow_test
    def test_pipe_through_stateconf(self, dirpath):
        output = os.path.join(dirpath, "output")
        write_to(
            os.path.join(dirpath, "xxx.sls"),
            textwrap.dedent(
                """#!stateconf -os yaml . jinja
            .X:
              cmd.run:
                - name: echo X >> {0}
                - cwd: /
            .Y:
              cmd.run:
                - name: echo Y >> {0}
                - cwd: /
            .Z:
              cmd.run:
                - name: echo Z >> {0}
                - cwd: /
            """.format(
                    output.replace("\\", "/")
                )
            ),
        )
        write_to(
            os.path.join(dirpath, "yyy.sls"),
            textwrap.dedent(
                """\
            #!pydsl|stateconf -ps

            __pydsl__.set(ordered=True)
            state('.D').cmd.run('echo D >> {0}', cwd='/')
            state('.E').cmd.run('echo E >> {0}', cwd='/')
            state('.F').cmd.run('echo F >> {0}', cwd='/')
            """.format(
                    output.replace("\\", "/")
                )
            ),
        )

        write_to(
            os.path.join(dirpath, "aaa.sls"),
            textwrap.dedent(
                """\
            #!pydsl|stateconf -ps

            include('xxx', 'yyy')

            # make all states in xxx run BEFORE states in this sls.
            extend(state('.start').stateconf.require(stateconf='xxx::goal'))

            # make all states in yyy run AFTER this sls.
            extend(state('.goal').stateconf.require_in(stateconf='yyy::start'))

            __pydsl__.set(ordered=True)

            state('.A').cmd.run('echo A >> {0}', cwd='/')
            state('.B').cmd.run('echo B >> {0}', cwd='/')
            state('.C').cmd.run('echo C >> {0}', cwd='/')
            """.format(
                    output.replace("\\", "/")
                )
            ),
        )

        self.state_highstate({"base": ["aaa"]}, dirpath)
        with salt.utils.files.fopen(output, "r") as f:
            self.assertEqual("".join(f.read().split()), "XYZABCDEF")

    @with_tempdir()
    @pytest.mark.slow_test
    def test_compile_time_state_execution(self, dirpath):
        if not sys.stdin.isatty():
            self.skipTest("Not attached to a TTY")
        # The Windows shell will include any spaces before the redirect
        # in the text that is redirected.
        # For example: echo hello > test.txt will contain "hello "
        write_to(
            os.path.join(dirpath, "aaa.sls"),
            textwrap.dedent(
                """\
            #!pydsl

            __pydsl__.set(ordered=True)
            A = state('A')
            A.cmd.run('echo hehe>{0}/zzz.txt', cwd='/')
            A.file.managed('{0}/yyy.txt', source='salt://zzz.txt')
            A()
            A()

            state().cmd.run('echo hoho>>{0}/yyy.txt', cwd='/')

            A.file.managed('{0}/xxx.txt', source='salt://zzz.txt')
            A()
            """.format(
                    dirpath.replace("\\", "/")
                )
            ),
        )
        self.state_highstate({"base": ["aaa"]}, dirpath)
        with salt.utils.files.fopen(os.path.join(dirpath, "yyy.txt"), "rt") as f:
            self.assertEqual(f.read(), "hehe" + os.linesep + "hoho" + os.linesep)
        with salt.utils.files.fopen(os.path.join(dirpath, "xxx.txt"), "rt") as f:
            self.assertEqual(f.read(), "hehe" + os.linesep)

    @with_tempdir()
    @pytest.mark.slow_test
    def test_nested_high_state_execution(self, dirpath):
        output = os.path.join(dirpath, "output")
        write_to(
            os.path.join(dirpath, "aaa.sls"),
            textwrap.dedent(
                """\
            #!pydsl
            __salt__['state.sls']('bbb')
            state().cmd.run('echo bbbbbb', cwd='/')
            """
            ),
        )
        write_to(
            os.path.join(dirpath, "bbb.sls"),
            textwrap.dedent(
                """
            # {{ salt['state.sls']('ccc') }}
            test:
              cmd.run:
                - name: echo bbbbbbb
                - cwd: /
            """
            ),
        )
        write_to(
            os.path.join(dirpath, "ccc.sls"),
            textwrap.dedent(
                """
            #!pydsl
            state().cmd.run('echo ccccc', cwd='/')
            """
            ),
        )
        self.state_highstate({"base": ["aaa"]}, dirpath)

    @with_tempdir()
    @pytest.mark.slow_test
    def test_repeat_includes(self, dirpath):
        output = os.path.join(dirpath, "output")
        write_to(
            os.path.join(dirpath, "b.sls"),
            textwrap.dedent(
                """\
            #!pydsl
            include('c')
            include('d')
            """
            ),
        )
        write_to(
            os.path.join(dirpath, "c.sls"),
            textwrap.dedent(
                """\
            #!pydsl
            modtest = include('e')
            modtest.success
            """
            ),
        )
        write_to(
            os.path.join(dirpath, "d.sls"),
            textwrap.dedent(
                """\
            #!pydsl
            modtest = include('e')
            modtest.success
            """
            ),
        )
        write_to(
            os.path.join(dirpath, "e.sls"),
            textwrap.dedent(
                """\
            #!pydsl
            success = True
            """
            ),
        )
        self.state_highstate({"base": ["b"]}, dirpath)
        self.state_highstate({"base": ["c", "d"]}, dirpath)


def write_to(fpath, content):
    with salt.utils.files.fopen(fpath, "w") as f:
        f.write(content)



"""
Verify salt-ssh passes on a failing retcode from state execution.
"""

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module", autouse=True)
def state_tree_run_fail(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_run
      '127.0.0.1':
        - fail_run
    """
    state_file = """
    This file state fails:
      file.managed:
        - name: /tmp/non/ex/is/tent
        - makedirs: false
        - contents: foo
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_run.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.mark.parametrize(
    "args",
    (
        ("state.sls", "fail_run"),
        ("state.highstate",),
        ("state.sls_id", "This file state fails", "fail_run"),
        ("state.top", "top.sls"),
    ),
)
def test_it(salt_ssh_cli, args):
    ret = salt_ssh_cli.run(*args)
    assert ret.returncode == EX_AGGREGATE



"""
    tests.e2e.compat.test_with_versions
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test current salt master with older salt minions
"""

import logging
import pathlib

import pytest
from saltfactories.daemons.container import SaltMinion
from saltfactories.utils import random_string

import salt.utils.platform
from tests.conftest import FIPS_TESTRUN
from tests.support.runtests import RUNTIME_VARS

docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker"),
    pytest.mark.skipif(
        salt.utils.platform.is_photonos() is True, reason="Skip on PhotonOS"
    ),
]


def _get_test_versions_ids(value):
    return f"SaltMinion~={value}"


@pytest.fixture(
    params=("3002", "3003", "3004"), ids=_get_test_versions_ids, scope="module"
)
def compat_salt_version(request):
    return request.param


@pytest.fixture(scope="module")
def minion_image_name(compat_salt_version):
    return f"salt-{compat_salt_version}"


@pytest.fixture(scope="function")
def minion_id(compat_salt_version):
    return random_string(
        f"salt-{compat_salt_version}-",
        uppercase=False,
    )


@pytest.fixture(scope="function")
def artifacts_path(minion_id, tmp_path):
    yield tmp_path / minion_id


@pytest.mark.skip_if_binaries_missing("docker")
@pytest.fixture(scope="function")
def salt_minion(
    minion_id,
    salt_master,
    docker_client,
    artifacts_path,
    compat_salt_version,
    host_docker_network_ip_address,
):
    config_overrides = {
        "master": salt_master.config["interface"],
        "user": False,
        "pytest-minion": {
            "log": {"host": host_docker_network_ip_address},
            "returner_address": {"host": host_docker_network_ip_address},
        },
        # We also want to scrutinize the key acceptance
        "open_mode": False,
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_master.salt_minion_daemon(
        minion_id,
        overrides=config_overrides,
        factory_class=SaltMinion,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
        # SaltMinion kwargs
        name=minion_id,
        image="ghcr.io/saltstack/salt-ci-containers/salt:{}".format(
            compat_salt_version
        ),
        docker_client=docker_client,
        start_timeout=120,
        pull_before_start=False,
        skip_if_docker_client_not_connectable=True,
        container_run_kwargs={
            "volumes": {
                str(artifacts_path): {
                    "bind": "/artifacts",
                    "mode": "z",
                },
            }
        },
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, salt_master, factory.id
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="function")
def package_name():
    return "figlet"


@pytest.fixture
def populated_state_tree(minion_id, package_name, state_tree):
    module_contents = """
    def get_test_package_name():
        return "{}"
    """.format(
        package_name
    )
    top_file_contents = """
    base:
        {}:
          - install-package
    """.format(
        minion_id
    )
    install_package_sls_contents = """
    state-entry-contém-unicode:
        pkg.installed:
          - name: {{ salt.pkgnames.get_test_package_name() }}
    """
    with pytest.helpers.temp_file(
        "_modules/pkgnames.py",
        module_contents,
        state_tree,
    ), pytest.helpers.temp_file(
        "top.sls", top_file_contents, state_tree
    ), pytest.helpers.temp_file(
        "install-package.sls",
        install_package_sls_contents,
        state_tree,
    ):
        # Run the test
        yield


@pytest.mark.skip_on_fips_enabled_platform
def test_ping(salt_cli, salt_minion):
    ret = salt_cli.run("test.ping", minion_tgt=salt_minion.id)
    assert ret.returncode == 0, ret
    assert ret.data is True


@pytest.mark.skip_on_fips_enabled_platform
@pytest.mark.usefixtures("populated_state_tree")
def test_highstate(salt_cli, salt_minion, package_name):
    """
    Assert a state.highstate with a newer master runs properly on older minions.
    """
    ret = salt_cli.run("state.highstate", minion_tgt=salt_minion.id, _timeout=300)
    assert ret.returncode == 0, ret
    assert ret.data is not None
    assert isinstance(ret.data, dict), ret.data
    state_return = next(iter(ret.data.values()))
    assert package_name in state_return["changes"], state_return


@pytest.mark.skip_on_fips_enabled_platform
@pytest.fixture
def cp_file_source():
    source = pathlib.Path(RUNTIME_VARS.BASE_FILES) / "cheese"
    contents = source.read_text().replace("ee", "æ")
    with pytest.helpers.temp_file(contents=contents) as temp_file:
        yield pathlib.Path(temp_file)


@pytest.mark.skip_on_fips_enabled_platform
def test_cp(salt_cp_cli, salt_minion, artifacts_path, cp_file_source):
    """
    Assert proper behaviour for salt-cp with a newer master and older minions.
    """
    remote_path = "/artifacts/cheese"
    ret = salt_cp_cli.run(
        str(cp_file_source), remote_path, minion_tgt=salt_minion.id, _timeout=300
    )
    assert ret.returncode == 0, ret
    assert ret.data is not None
    assert isinstance(ret.data, dict), ret.data
    assert ret.data == {remote_path: True}
    cp_file_dest = artifacts_path / "cheese"
    assert cp_file_source.read_text() == cp_file_dest.read_text()



import pathlib

import pytest

import salt.modules.win_file as file
import salt.modules.win_lgpo_reg as win_lgpo_reg
import salt.states.win_lgpo_reg as lgpo_reg
import salt.utils.files
import salt.utils.win_dacl
import salt.utils.win_lgpo_reg
import salt.utils.win_reg
from tests.support.mock import patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
]


@pytest.fixture
def configure_loader_modules():
    return {
        lgpo_reg: {
            "__opts__": {"test": False},
            "__salt__": {
                "lgpo_reg.get_value": win_lgpo_reg.get_value,
                "lgpo_reg.set_value": win_lgpo_reg.set_value,
                "lgpo_reg.disable_value": win_lgpo_reg.disable_value,
                "lgpo_reg.delete_value": win_lgpo_reg.delete_value,
            },
            "__utils__": {
                "reg.read_value": salt.utils.win_reg.read_value,
            },
        },
        file: {
            "__utils__": {
                "dacl.set_perms": salt.utils.win_dacl.set_perms,
            },
        },
    }


@pytest.fixture
def empty_reg_pol_mach():
    class_info = salt.utils.win_lgpo_reg.CLASS_INFO
    reg_pol_file = pathlib.Path(class_info["Machine"]["policy_path"])
    reg_pol_file.parent.mkdir(parents=True, exist_ok=True)
    reg_pol_file.write_bytes(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey2")
    yield
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey2")
    reg_pol_file.write_bytes(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))


@pytest.fixture
def empty_reg_pol_user():
    class_info = salt.utils.win_lgpo_reg.CLASS_INFO
    reg_pol_file = pathlib.Path(class_info["User"]["policy_path"])
    reg_pol_file.parent.mkdir(parents=True, exist_ok=True)
    reg_pol_file.write_bytes(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey2")
    yield
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey2")
    reg_pol_file.write_bytes(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))


@pytest.fixture
def reg_pol_mach():
    data_to_write = {
        r"SOFTWARE\MyKey1": {
            "MyValue1": {
                "data": "squidward",
                "type": "REG_SZ",
            },
            "**del.MyValue2": {
                "data": " ",
                "type": "REG_SZ",
            },
            "MyValue3": {
                "data": 0,
                "type": "REG_DWORD",
            },
        },
        r"SOFTWARE\MyKey2": {
            "MyValue3": {
                "data": ["spongebob", "squarepants"],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    win_lgpo_reg.write_reg_pol(data_to_write)
    salt.utils.win_reg.set_value(
        hive="HKLM",
        key="SOFTWARE\\MyKey1",
        vname="MyValue1",
        vdata="squidward",
        vtype="REG_SZ",
    )
    salt.utils.win_reg.set_value(
        hive="HKLM",
        key="SOFTWARE\\MyKey1",
        vname="MyValue3",
        vdata=0,
        vtype="REG_DWORD",
    )
    salt.utils.win_reg.set_value(
        hive="HKLM",
        key="SOFTWARE\\MyKey2",
        vname="MyValue3",
        vdata=["spongebob", "squarepants"],
        vtype="REG_MULTI_SZ",
    )
    yield
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKLM", key="SOFTWARE\\MyKey2")
    class_info = salt.utils.win_lgpo_reg.CLASS_INFO
    reg_pol_file = class_info["Machine"]["policy_path"]
    with salt.utils.files.fopen(reg_pol_file, "wb") as f:
        f.write(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))


@pytest.fixture
def reg_pol_user():
    data_to_write = {
        r"SOFTWARE\MyKey1": {
            "MyValue1": {
                "data": "squidward",
                "type": "REG_SZ",
            },
            "**del.MyValue2": {
                "data": " ",
                "type": "REG_SZ",
            },
            "MyValue3": {
                "data": 0,
                "type": "REG_DWORD",
            },
        },
        r"SOFTWARE\MyKey2": {
            "MyValue3": {
                "data": ["spongebob", "squarepants"],
                "type": "REG_MULTI_SZ",
            },
        },
    }
    win_lgpo_reg.write_reg_pol(data_to_write, policy_class="User")
    salt.utils.win_reg.set_value(
        hive="HKCU",
        key="SOFTWARE\\MyKey1",
        vname="MyValue1",
        vdata="squidward",
        vtype="REG_SZ",
    )
    salt.utils.win_reg.set_value(
        hive="HKCU",
        key="SOFTWARE\\MyKey1",
        vname="MyValue3",
        vdata=0,
        vtype="REG_DWORD",
    )
    salt.utils.win_reg.set_value(
        hive="HKCU",
        key="SOFTWARE\\MyKey2",
        vname="MyValue3",
        vdata=["spongebob", "squarepants"],
        vtype="REG_MULTI_SZ",
    )
    yield
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey1")
    salt.utils.win_reg.delete_key_recursive(hive="HKCU", key="SOFTWARE\\MyKey2")
    class_info = salt.utils.win_lgpo_reg.CLASS_INFO
    reg_pol_file = class_info["User"]["policy_path"]
    with salt.utils.files.fopen(reg_pol_file, "wb") as f:
        f.write(salt.utils.win_lgpo_reg.REG_POL_HEADER.encode("utf-16-le"))


def test_virtual_name():
    assert lgpo_reg.__virtual__() == "lgpo_reg"


def test_mach_value_present(empty_reg_pol_mach):
    """
    Test value.present in Machine policy
    """
    result = lgpo_reg.value_present(
        name="MyValue",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue",
                    "data": 1,
                    "type": "REG_DWORD",
                },
                "reg": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue",
                    "data": 1,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {},
                "reg": {},
            },
        },
        "comment": "Policy value set\nRegistry value set",
        "name": "MyValue",
        "result": True,
    }
    assert result == expected


def test_mach_value_present_similar_names(empty_reg_pol_mach):
    """
    Test value.present in Machine policy
    """
    lgpo_reg.value_present(
        name="MyValueTwo",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
    )
    lgpo_reg.value_present(
        name="MyValue",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
    )
    expected = {
        "SOFTWARE\\MyKey1": {
            "MyValue": {
                "type": "REG_DWORD",
                "data": 1,
            },
            "MyValueTwo": {
                "type": "REG_DWORD",
                "data": 1,
            },
        },
    }
    result = win_lgpo_reg.read_reg_pol(policy_class="Machine")
    assert result == expected


def test_mach_value_present_enforce(reg_pol_mach):
    """
    Issue #64222
    Test value.present in Machine policy when the registry changes after the
    state is applied. This would cause a discrepancy between the registry
    setting and the value in the registry.pol file
    """
    # reg_pol_mach has MyValue3 with REG_DWORD value of 0, let's set it to 1
    salt.utils.win_reg.set_value(
        hive="HKLM",
        key="SOFTWARE\\MyKey1",
        vname="MyValue3",
        vdata="1",
        vtype="REG_DWORD",
    )
    # Now the registry and Registry.pol file are out of sync
    result = lgpo_reg.value_present(
        name="MyValue3",
        key="SOFTWARE\\MyKey1",
        v_data="0",
        v_type="REG_DWORD",
    )
    expected = {
        "changes": {
            "new": {
                "reg": {
                    "data": 0,
                }
            },
            "old": {
                "reg": {
                    "data": 1,
                }
            },
        },
        "comment": "Registry value set",
        "name": "MyValue3",
        "result": True,
    }
    assert result == expected


def test_mach_value_present_existing_change(reg_pol_mach):
    """
    Test value.present with existing incorrect value in Machine policy
    """
    result = lgpo_reg.value_present(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        v_data="2",
        v_type="REG_DWORD",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 2,
                    "type": "REG_DWORD",
                },
                "reg": {
                    "data": 2,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {
                    "data": "squidward",
                    "type": "REG_SZ",
                },
                "reg": {
                    "data": "squidward",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value set\nRegistry value set",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_present_existing_change_dword(reg_pol_mach):
    """
    Test value.present with existing incorrect value in Machine policy
    """
    result = lgpo_reg.value_present(
        name="MyValue3",
        key="SOFTWARE\\MyKey1",
        v_data=1,
        v_type="REG_DWORD",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 1,
                },
                "reg": {
                    "data": 1,
                },
            },
            "old": {
                "pol": {
                    "data": 0,
                },
                "reg": {
                    "data": 0,
                },
            },
        },
        "comment": "Policy value set\nRegistry value set",
        "name": "MyValue3",
        "result": True,
    }
    assert result == expected


def test_mach_value_present_existing_no_change(reg_pol_mach):
    """
    Test value.present with existing correct value in Machine policy
    """
    result = lgpo_reg.value_present(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        v_data="squidward",
        v_type="REG_SZ",
    )
    expected = {
        "changes": {},
        "comment": "Policy value already present\nRegistry value already present",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_present_test_true(empty_reg_pol_mach):
    """
    Test value.present with test=True in Machine policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_present(
            name="MyValue",
            key="SOFTWARE\\MyKey1",
            v_data="1",
            v_type="REG_DWORD",
        )
    expected = {
        "changes": {},
        "comment": "Policy value will be set\nRegistry value will be set",
        "name": "MyValue",
        "result": None,
    }
    assert result == expected


def test_mach_value_present_existing_disabled(reg_pol_mach):
    """
    Test value.present with existing value that is disabled in Machine policy
    """
    result = lgpo_reg.value_present(
        name="MyValue2",
        key="SOFTWARE\\MyKey1",
        v_data="2",
        v_type="REG_DWORD",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 2,
                    "type": "REG_DWORD",
                },
                "reg": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue2",
                    "data": 2,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {
                    "data": "**del.MyValue2",
                    "type": "REG_SZ",
                },
                "reg": {},
            },
        },
        "comment": "Policy value set\nRegistry value set",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_mach_value_disabled(empty_reg_pol_mach):
    """
    Test value.disabled in Machine policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "**del.MyValue1",
                    "type": "REG_SZ",
                },
            },
            "old": {"pol": {}},
        },
        "comment": "Policy value disabled",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_disabled_existing_change(reg_pol_mach):
    """
    Test value.disabled with an existing value that is not disabled in Machine
    policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": "**del.MyValue1",
                },
                "reg": {},
            },
            "old": {
                "pol": {
                    "data": "squidward",
                },
                "reg": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "squidward",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value disabled\nRegistry value deleted",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_disabled_existing_no_change(reg_pol_mach):
    """
    Test value.disabled with an existing disabled value in Machine policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue2",
        key="SOFTWARE\\MyKey1",
    )
    expected = {
        "changes": {},
        "comment": "Policy value already disabled\nRegistry value already deleted",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_mach_value_disabled_test_true(empty_reg_pol_mach):
    """
    Test value.disabled when test=True in Machine policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_disabled(
            name="MyValue",
            key="SOFTWARE\\MyKey1",
        )
    expected = {
        "changes": {},
        "comment": "Policy value will be disabled",
        "name": "MyValue",
        "result": None,
    }
    assert result == expected


def test_mach_value_absent(reg_pol_mach):
    """
    Test value.absent in Machine policy
    """
    result = lgpo_reg.value_absent(name="MyValue1", key="SOFTWARE\\MyKey1")
    expected = {
        "changes": {
            "new": {"pol": {}, "reg": {}},
            "old": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "squidward",
                    "type": "REG_SZ",
                },
                "reg": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "squidward",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value deleted\nRegistry value deleted",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_absent_no_change(empty_reg_pol_mach):
    """
    Test value.absent when the value is already absent in Machine policy
    """
    result = lgpo_reg.value_absent(name="MyValue1", key="SOFTWARE\\MyKey1")
    expected = {
        "changes": {},
        "comment": "Policy value already deleted\nRegistry value already deleted",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_mach_value_absent_disabled(reg_pol_mach):
    """
    Test value.absent when the value is disabled in Machine policy
    """
    result = lgpo_reg.value_absent(name="MyValue2", key="SOFTWARE\\MyKey1")
    expected = {
        "changes": {
            "new": {"pol": {}},
            "old": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue2",
                    "data": "**del.MyValue2",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value deleted",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_mach_value_absent_test_true(reg_pol_mach):
    """
    Test value.absent with test=True in Machine policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_absent(name="MyValue1", key="SOFTWARE\\MyKey1")
    expected = {
        "changes": {},
        "comment": "Policy value will be deleted\nRegistry value will be deleted",
        "name": "MyValue1",
        "result": None,
    }
    assert result == expected


def test_user_value_present(empty_reg_pol_user):
    """
    Test value.present in User policy
    """
    result = lgpo_reg.value_present(
        name="MyValue",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue",
                    "data": 1,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {},
            },
        },
        "comment": "Policy value set",
        "name": "MyValue",
        "result": True,
    }
    assert result == expected


def test_user_value_present_similar_names(empty_reg_pol_user):
    """
    Test value.present in User policy
    """
    lgpo_reg.value_present(
        name="MyValueTwo",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
        policy_class="User",
    )
    lgpo_reg.value_present(
        name="MyValue",
        key="SOFTWARE\\MyKey1",
        v_data="1",
        v_type="REG_DWORD",
        policy_class="User",
    )
    expected = {
        "SOFTWARE\\MyKey1": {
            "MyValue": {
                "type": "REG_DWORD",
                "data": 1,
            },
            "MyValueTwo": {
                "type": "REG_DWORD",
                "data": 1,
            },
        },
    }
    result = win_lgpo_reg.read_reg_pol(policy_class="User")
    assert result == expected


def test_user_value_present_existing_change(reg_pol_user):
    """
    Test value.present with existing incorrect value in User policy
    """
    result = lgpo_reg.value_present(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        v_data="2",
        v_type="REG_DWORD",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 2,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {
                    "data": "squidward",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value set",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_present_existing_change_dword(reg_pol_user):
    """
    Test value.present with existing incorrect value in User policy
    """
    result = lgpo_reg.value_present(
        name="MyValue3",
        key="SOFTWARE\\MyKey1",
        v_data=1,
        v_type="REG_DWORD",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 1,
                },
            },
            "old": {
                "pol": {
                    "data": 0,
                },
            },
        },
        "comment": "Policy value set",
        "name": "MyValue3",
        "result": True,
    }
    assert result == expected


def test_user_value_present_existing_no_change(reg_pol_user):
    """
    Test value.present with existing correct value in User policy
    """
    result = lgpo_reg.value_present(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        v_data="squidward",
        v_type="REG_SZ",
        policy_class="User",
    )
    expected = {
        "changes": {},
        "comment": "Policy value already present",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_present_test_true(empty_reg_pol_user):
    """
    Test value.present with test=True in User policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_present(
            name="MyValue",
            key="SOFTWARE\\MyKey1",
            v_data="1",
            v_type="REG_DWORD",
            policy_class="User",
        )
    expected = {
        "changes": {},
        "comment": "Policy value will be set",
        "name": "MyValue",
        "result": None,
    }
    assert result == expected


def test_user_value_present_existing_disabled(reg_pol_user):
    """
    Test value.present with existing value that is disabled in User policy
    """
    result = lgpo_reg.value_present(
        name="MyValue2",
        key="SOFTWARE\\MyKey1",
        v_data="2",
        v_type="REG_DWORD",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": 2,
                    "type": "REG_DWORD",
                },
            },
            "old": {
                "pol": {
                    "data": "**del.MyValue2",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value set",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_user_value_disabled(empty_reg_pol_user):
    """
    Test value.disabled in User policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue1", key="SOFTWARE\\MyKey1", policy_class="User"
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "**del.MyValue1",
                    "type": "REG_SZ",
                },
            },
            "old": {"pol": {}},
        },
        "comment": "Policy value disabled",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_disabled_existing_change(reg_pol_user):
    """
    Test value.disabled with an existing value that is not disabled in User
    policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {
                    "data": "**del.MyValue1",
                },
            },
            "old": {
                "pol": {
                    "data": "squidward",
                },
            },
        },
        "comment": "Policy value disabled",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_disabled_existing_no_change(reg_pol_user):
    """
    Test value.disabled with an existing disabled value in User policy
    """
    result = lgpo_reg.value_disabled(
        name="MyValue2",
        key="SOFTWARE\\MyKey1",
        policy_class="User",
    )
    expected = {
        "changes": {},
        "comment": "Policy value already disabled",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_user_value_disabled_test_true(empty_reg_pol_user):
    """
    Test value.disabled when test=True in User policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_disabled(
            name="MyValue",
            key="SOFTWARE\\MyKey1",
            policy_class="User",
        )
    expected = {
        "changes": {},
        "comment": "Policy value will be disabled",
        "name": "MyValue",
        "result": None,
    }
    assert result == expected


def test_user_value_absent(reg_pol_user):
    """
    Test value.absent in User policy
    """
    result = lgpo_reg.value_absent(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {
                "pol": {},
            },
            "old": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue1",
                    "data": "squidward",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value deleted",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_absent_no_change(empty_reg_pol_user):
    """
    Test value.absent when the value is already absent in User policy
    """
    result = lgpo_reg.value_absent(
        name="MyValue1",
        key="SOFTWARE\\MyKey1",
        policy_class="User",
    )
    expected = {
        "changes": {},
        "comment": "Policy value already deleted",
        "name": "MyValue1",
        "result": True,
    }
    assert result == expected


def test_user_value_absent_disabled(reg_pol_user):
    """
    Test value.absent when the value is disabled in User policy
    """
    result = lgpo_reg.value_absent(
        name="MyValue2",
        key="SOFTWARE\\MyKey1",
        policy_class="User",
    )
    expected = {
        "changes": {
            "new": {"pol": {}},
            "old": {
                "pol": {
                    "key": "SOFTWARE\\MyKey1",
                    "name": "MyValue2",
                    "data": "**del.MyValue2",
                    "type": "REG_SZ",
                },
            },
        },
        "comment": "Policy value deleted",
        "name": "MyValue2",
        "result": True,
    }
    assert result == expected


def test_user_value_absent_test_true(reg_pol_user):
    """
    Test value.absent with test=True in User policy
    """
    with patch.dict(lgpo_reg.__opts__, {"test": True}):
        result = lgpo_reg.value_absent(
            name="MyValue1",
            key="SOFTWARE\\MyKey1",
            policy_class="User",
        )
    expected = {
        "changes": {},
        "comment": "Policy value will be deleted",
        "name": "MyValue1",
        "result": None,
    }
    assert result == expected



import logging

import pytest

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
]

RPM_GPG_KEY_EPEL_8_SALTTEST = """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mQINBFz3zvsBEADJOIIWllGudxnpvJnkxQz2CtoWI7godVnoclrdl83kVjqSQp+2
dgxuG5mUiADUfYHaRQzxKw8efuQnwxzU9kZ70ngCxtmbQWGmUmfSThiapOz00018
+eo5MFabd2vdiGo1y+51m2sRDpN8qdCaqXko65cyMuLXrojJHIuvRA/x7iqOrRfy
a8x3OxC4PEgl5pgDnP8pVK0lLYncDEQCN76D9ubhZQWhISF/zJI+e806V71hzfyL
/Mt3mQm/li+lRKU25Usk9dWaf4NH/wZHMIPAkVJ4uD4H/uS49wqWnyiTYGT7hUbi
ecF7crhLCmlRzvJR8mkRP6/4T/F3tNDPWZeDNEDVFUkTFHNU6/h2+O398MNY/fOh
yKaNK3nnE0g6QJ1dOH31lXHARlpFOtWt3VmZU0JnWLeYdvap4Eff9qTWZJhI7Cq0
Wm8DgLUpXgNlkmquvE7P2W5EAr2E5AqKQoDbfw/GiWdRvHWKeNGMRLnGI3QuoX3U
pAlXD7v13VdZxNydvpeypbf/AfRyrHRKhkUj3cU1pYkM3DNZE77C5JUe6/0nxbt4
ETUZBTgLgYJGP8c7PbkVnO6I/KgL1jw+7MW6Az8Ox+RXZLyGMVmbW/TMc8haJfKL
MoUo3TVk8nPiUhoOC0/kI7j9ilFrBxBU5dUtF4ITAWc8xnG6jJs/IsvRpQARAQAB
tChGZWRvcmEgRVBFTCAoOCkgPGVwZWxAZmVkb3JhcHJvamVjdC5vcmc+iQI4BBMB
AgAiBQJc9877AhsPBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAKCRAh6kWrL4bW
oWagD/4xnLWws34GByVDQkjprk0fX7Iyhpm/U7BsIHKspHLL+Y46vAAGY/9vMvdE
0fcr9Ek2Zp7zE1RWmSCzzzUgTG6BFoTG1H4Fho/7Z8BXK/jybowXSZfqXnTOfhSF
alwDdwlSJvfYNV9MbyvbxN8qZRU1z7PEWZrIzFDDToFRk0R71zHpnPTNIJ5/YXTw
NqU9OxII8hMQj4ufF11040AJQZ7br3rzerlyBOB+Jd1zSPVrAPpeMyJppWFHSDAI
WK6x+am13VIInXtqB/Cz4GBHLFK5d2/IYspVw47Solj8jiFEtnAq6+1Aq5WH3iB4
bE2e6z00DSF93frwOyWN7WmPIoc2QsNRJhgfJC+isGQAwwq8xAbHEBeuyMG8GZjz
xohg0H4bOSEujVLTjH1xbAG4DnhWO/1VXLX+LXELycO8ZQTcjj/4AQKuo4wvMPrv
9A169oETG+VwQlNd74VBPGCvhnzwGXNbTK/KH1+WRH0YSb+41flB3NKhMSU6dGI0
SGtIxDSHhVVNmx2/6XiT9U/znrZsG5Kw8nIbbFz+9MGUUWgJMsd1Zl9R8gz7V9fp
n7L7y5LhJ8HOCMsY/Z7/7HUs+t/A1MI4g7Q5g5UuSZdgi0zxukiWuCkLeAiAP4y7
zKK4OjJ644NDcWCHa36znwVmkz3ixL8Q0auR15Oqq2BjR/fyog==
=84m8
-----END PGP PUBLIC KEY BLOCK-----
"""

RPM_GPG_KEY_EPEL_7_SALTTEST = """\
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1.4.11 (GNU/Linux)

mQINBFKuaIQBEAC1UphXwMqCAarPUH/ZsOFslabeTVO2pDk5YnO96f+rgZB7xArB
OSeQk7B90iqSJ85/c72OAn4OXYvT63gfCeXpJs5M7emXkPsNQWWSju99lW+AqSNm
jYWhmRlLRGl0OO7gIwj776dIXvcMNFlzSPj00N2xAqjMbjlnV2n2abAE5gq6VpqP
vFXVyfrVa/ualogDVmf6h2t4Rdpifq8qTHsHFU3xpCz+T6/dGWKGQ42ZQfTaLnDM
jToAsmY0AyevkIbX6iZVtzGvanYpPcWW4X0RDPcpqfFNZk643xI4lsZ+Y2Er9Yu5
S/8x0ly+tmmIokaE0wwbdUu740YTZjCesroYWiRg5zuQ2xfKxJoV5E+Eh+tYwGDJ
n6HfWhRgnudRRwvuJ45ztYVtKulKw8QQpd2STWrcQQDJaRWmnMooX/PATTjCBExB
9dkz38Druvk7IkHMtsIqlkAOQMdsX1d3Tov6BE2XDjIG0zFxLduJGbVwc/6rIc95
T055j36Ez0HrjxdpTGOOHxRqMK5m9flFbaxxtDnS7w77WqzW7HjFrD0VeTx2vnjj
GqchHEQpfDpFOzb8LTFhgYidyRNUflQY35WLOzLNV+pV3eQ3Jg11UFwelSNLqfQf
uFRGc+zcwkNjHh5yPvm9odR1BIfqJ6sKGPGbtPNXo7ERMRypWyRz0zi0twARAQAB
tChGZWRvcmEgRVBFTCAoNykgPGVwZWxAZmVkb3JhcHJvamVjdC5vcmc+iQI4BBMB
AgAiBQJSrmiEAhsPBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAKCRBqL66iNSxk
5cfGD/4spqpsTjtDM7qpytKLHKruZtvuWiqt5RfvT9ww9GUUFMZ4ZZGX4nUXg49q
ixDLayWR8ddG/s5kyOi3C0uX/6inzaYyRg+Bh70brqKUK14F1BrrPi29eaKfG+Gu
MFtXdBG2a7OtPmw3yuKmq9Epv6B0mP6E5KSdvSRSqJWtGcA6wRS/wDzXJENHp5re
9Ism3CYydpy0GLRA5wo4fPB5uLdUhLEUDvh2KK//fMjja3o0L+SNz8N0aDZyn5Ax
CU9RB3EHcTecFgoy5umRj99BZrebR1NO+4gBrivIfdvD4fJNfNBHXwhSH9ACGCNv
HnXVjHQF9iHWApKkRIeh8Fr2n5dtfJEF7SEX8GbX7FbsWo29kXMrVgNqHNyDnfAB
VoPubgQdtJZJkVZAkaHrMu8AytwT62Q4eNqmJI1aWbZQNI5jWYqc6RKuCK6/F99q
thFT9gJO17+yRuL6Uv2/vgzVR1RGdwVLKwlUjGPAjYflpCQwWMAASxiv9uPyYPHc
ErSrbRG0wjIfAR3vus1OSOx3xZHZpXFfmQTsDP7zVROLzV98R3JwFAxJ4/xqeON4
vCPFU6OsT3lWQ8w7il5ohY95wmujfr6lk89kEzJdOTzcn7DBbUru33CQMGKZ3Evt
RjsC7FDbL017qxS+ZVA/HGkyfiu4cpgV8VUnbql5eAZ+1Ll6Dw==
=hdPa
-----END PGP PUBLIC KEY BLOCK-----
"""


@pytest.fixture
def pkgrepo(states, grains):
    if grains["os_family"] != "RedHat":
        raise pytest.skip.Exception(
            "Test only for CentOS platforms, not '{}' based distributions.".format(
                grains["os_family"]
            ),
            _use_item_location=True,
        )
    return states.pkgrepo


@pytest.fixture
def centos_state_tree(grains, pkgrepo, state_tree):
    if grains["os"] not in ("CentOS", "CentOS Stream"):
        pytest.skip("Test only applicable to CentOS, not '{}'.".format(grains["os"]))

    managed_sls_contents = """
    {% if grains['osmajorrelease'] == 8 %}
    epel-salttest:
      pkgrepo.managed:
        - humanname: Extra Packages for Enterprise Linux 8 - $basearch (salttest)
        - comments:
          - '#baseurl=http://download.fedoraproject.org/pub/epel/8/$basearch'
        - mirrorlist: https://mirrors.fedoraproject.org/metalink?repo=epel-8&arch=$basearch
        - failovermethod: priority
        - enabled: 1
        - gpgcheck: 1
        - gpgkey: file:///etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-8-salttest
        - require:
          - file: /etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-8-salttest

    /etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-8-salttest:
      file.managed:
        - source: salt://pkgrepo/files/RPM-GPG-KEY-EPEL-8-salttest
        - user: root
        - group: root
        - mode: 644
    {% elif grains['osmajorrelease'] == 7 %}
    epel-salttest:
      pkgrepo.managed:
        - humanname: Extra Packages for Enterprise Linux 7 - $basearch (salttest)
        - comments:
          - '#baseurl=http://download.fedoraproject.org/pub/epel/7/$basearch'
        - mirrorlist: https://mirrors.fedoraproject.org/metalink?repo=epel-7&arch=$basearch
        - failovermethod: priority
        - enabled: 1
        - gpgcheck: 1
        - gpgkey: file:///etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-7-salttest
        - require:
          - file: /etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-7-salttest

    /etc/pki/rpm-gpg/RPM-GPG-KEY-EPEL-7-salttest:
      file.managed:
        - source: salt://pkgrepo/files/RPM-GPG-KEY-EPEL-7-salttest
        - user: root
        - group: root
        - mode: 644
    {% endif %}
    """
    absend_sls_contents = """
    epel-salttest:
      pkgrepo:
        - absent
    """
    centos_7_gpg_key = pytest.helpers.temp_file(
        "pkgrepo/files/RPM-GPG-KEY-EPEL-7-salttest",
        RPM_GPG_KEY_EPEL_7_SALTTEST,
        state_tree,
    )
    centos_8_gpg_key = pytest.helpers.temp_file(
        "pkgrepo/files/RPM-GPG-KEY-EPEL-8-salttest",
        RPM_GPG_KEY_EPEL_8_SALTTEST,
        state_tree,
    )
    managed_state_file = pytest.helpers.temp_file(
        "pkgrepo/managed.sls", managed_sls_contents, state_tree
    )
    absent_state_file = pytest.helpers.temp_file(
        "pkgrepo/absent.sls", absend_sls_contents, state_tree
    )

    try:
        with centos_7_gpg_key, centos_8_gpg_key, managed_state_file, absent_state_file:
            yield
    finally:
        pass


@pytest.mark.requires_salt_states("pkgrepo.managed", "pkgrepo.absent")
def test_pkgrepo_managed_absent(grains, modules, subtests, centos_state_tree):
    """
    Test adding/removing a repo
    """
    if grains["os"] == "Fedora" and grains["osmajorrelease"] == 40:
        pytest.skip("No repo for {} in test COPR yet".format(grains["osfinger"]))

    add_repo_test_passed = False

    with subtests.test("Add Repo"):
        ret = modules.state.sls("pkgrepo.managed")
        assert ret.failed is False
        for state in ret:
            assert state.result is True
        add_repo_test_passed = True

    with subtests.test("Remove Repo"):
        if add_repo_test_passed is False:
            pytest.skip("Adding the repo failed. Skipping.")
        ret = modules.state.sls("pkgrepo.absent")
        assert ret.failed is False
        for state in ret:
            assert state.result is True


@pytest.fixture
def pkgrepo_with_comments_name(pkgrepo):
    pkgrepo_name = "examplerepo"
    try:
        yield pkgrepo_name
    finally:
        try:
            pkgrepo.absent(pkgrepo_name)
        except Exception:  # pylint: disable=broad-except
            pass


def test_pkgrepo_with_comments(grains, pkgrepo, pkgrepo_with_comments_name, subtests):
    """
    Test adding a repo with comments
    """
    if grains["os"] == "Fedora" and grains["osmajorrelease"] == 40:
        pytest.skip("No repo for {} in test COPR yet".format(grains["osfinger"]))

    kwargs = {
        "name": pkgrepo_with_comments_name,
        "baseurl": "http://example.com/repo",
        "enabled": False,
        "comments": ["This is a comment"],
    }

    with subtests.test("Add repo"):
        # Run the state to add the repo
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True

    with subtests.test("Modify comments"):
        # Run again with modified comments
        kwargs["comments"].append("This is another comment")
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True
        assert ret.changes == {
            "comments": {
                "old": ["This is a comment"],
                "new": ["This is a comment", "This is another comment"],
            }
        }

    with subtests.test("Repeat last call"):
        # Run a third time, no changes should be made
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True
        assert not ret.changes
        assert ret.comment == "Package repo '{}' already configured".format(
            pkgrepo_with_comments_name
        )


@pytest.fixture
def copr_pkgrepo_with_comments_name(pkgrepo, grains):
    if (
        grains["osfinger"] in ("CentOS Linux-7", "Amazon Linux-2")
        or grains["os"] == "VMware Photon OS"
    ):
        pytest.skip("copr plugin not installed on {} CI".format(grains["osfinger"]))
    if (
        grains["os"] in ("CentOS Stream", "AlmaLinux", "Rocky")
        and grains["osmajorrelease"] == 9
        or grains["osfinger"] == "Amazon Linux-2023"
    ):
        pytest.skip("No repo for {} in test COPR yet".format(grains["osfinger"]))

    pkgrepo_name = "hello-copr"
    try:
        yield pkgrepo_name
    finally:
        try:
            pkgrepo.absent(copr="mymindstorm/hello")
        except Exception:  # pylint: disable=broad-except
            pass


def test_copr_pkgrepo_with_comments(
    grains, pkgrepo, copr_pkgrepo_with_comments_name, subtests
):
    """
    Test adding a repo with comments
    """
    if grains["os"] == "Fedora" and grains["osmajorrelease"] == 40:
        pytest.skip("No repo for {} in test COPR yet".format(grains["osfinger"]))

    kwargs = {
        "name": copr_pkgrepo_with_comments_name,
        "copr": "mymindstorm/hello",
        "enabled": False,
        "comments": ["This is a comment"],
    }

    with subtests.test("Add repo"):
        # Run the state to add the repo
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True

    with subtests.test("Modify comments"):
        # Run again with modified comments
        kwargs["comments"].append("This is another comment")
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True
        assert ret.changes == {
            "comments": {
                "old": ["This is a comment"],
                "new": ["This is a comment", "This is another comment"],
            }
        }

    with subtests.test("Repeat last call"):
        # Run a third time, no changes should be made
        ret = pkgrepo.managed(**kwargs.copy())
        assert ret.result is True
        assert not ret.changes
        assert ret.comment == "Package repo '{}' already configured".format(
            copr_pkgrepo_with_comments_name
        )



"""
    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>
"""

import pytest

import salt.modules.napalm_network as napalm_network
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "__salt__": {
            "config.get": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "file.file_exists": napalm_test_support.true,
            "file.join": napalm_test_support.join,
            "file.get_managed": napalm_test_support.get_managed_file,
            "random.hash": napalm_test_support.random_hash,
        }
    }

    return {napalm_network: module_globals}


def test_connected_pass():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.connected()
        assert ret["out"] is True


def test_facts():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.facts()
        assert ret["out"] == napalm_test_support.TEST_FACTS.copy()


def test_environment():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.environment()
        assert ret["out"] == napalm_test_support.TEST_ENVIRONMENT.copy()


def test_cli_single_command():
    """
    Test that CLI works with 1 arg
    """
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.cli("show run")
        assert ret["out"] == napalm_test_support.TEST_COMMAND_RESPONSE.copy()


def test_cli_multi_command():
    """
    Test that CLI works with 2 arg
    """
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.cli("show run", "show run")
        assert ret["out"] == napalm_test_support.TEST_COMMAND_RESPONSE.copy()


def test_traceroute():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.traceroute("destination.com")
        assert list(ret["out"].keys())[0] == "success"


def test_ping():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.ping("destination.com")
        assert list(ret["out"].keys())[0] == "success"


def test_arp():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.arp()
        assert ret["out"] == napalm_test_support.TEST_ARP_TABLE.copy()


def test_ipaddrs():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.ipaddrs()
        assert ret["out"] == napalm_test_support.TEST_IPADDRS.copy()


def test_interfaces():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.interfaces()
        assert ret["out"] == napalm_test_support.TEST_INTERFACES.copy()


def test_lldp():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.lldp()
        assert ret["out"] == napalm_test_support.TEST_LLDP_NEIGHBORS.copy()


def test_mac():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.mac()
        assert ret["out"] == napalm_test_support.TEST_MAC_TABLE.copy()


def test_config():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.config("running")
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()


def test_optics():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.optics()
        assert ret["out"] == napalm_test_support.TEST_OPTICS.copy()


def test_load_config():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.load_config(text="new config")
        assert ret["result"]


def test_load_config_replace():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.load_config(text="new config", replace=True)
        assert ret["result"]


def test_load_template():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.load_template("set_ntp_peers", peers=["192.168.0.1"])
        assert ret["out"] is None


def test_commit():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.commit()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()


def test_discard_config():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.discard_config()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()


def test_compare_config():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.compare_config()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()


def test_rollback():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.rollback()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()


def test_config_changed():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.config_changed()
        assert ret == (True, "")


def test_config_control():
    with patch(
        "salt.utils.napalm.get_device",
        MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
    ):
        ret = napalm_network.config_control()
        assert ret == (True, "")



"""
Apache Libcloud Load Balancer State
===================================

Manage load balancers using libcloud

    :codeauthor: ``Anthony Shaw <anthonyshaw@apache.org>``

Apache Libcloud load balancer management for a full list
of supported clouds, see http://libcloud.readthedocs.io/en/latest/loadbalancer/supported_providers.html

Clouds include Amazon ELB, ALB, Google, Aliyun, CloudStack, Softlayer

.. versionadded:: 2018.3.0

:configuration:
    This module uses a configuration profile for one or multiple Cloud providers

    .. code-block:: yaml

        libcloud_loadbalancer:
            profile_test1:
              driver: gce
              key: GOOG0123456789ABCXYZ
              secret: mysecret
            profile_test2:
              driver: alb
              key: 12345
              secret: mysecret

Example:

Using States to deploy a load balancer with extended arguments to specify region

.. code-block:: yaml

    lb_test:
        libcloud_loadbalancer.balancer_present:
            - name: example
            - port: 80
            - protocol: http
            - profile: google
            - ex_region: us-east1

:depends: apache-libcloud
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    if "libcloud_loadbalancer.list_balancers" in __salt__:
        return True
    return (False, "libcloud_loadbalancer module could not be loaded")


def state_result(result, message, name, changes=None):
    if changes is None:
        changes = {}
    return {"result": result, "comment": message, "name": name, "changes": changes}


def balancer_present(
    name, port, protocol, profile, algorithm=None, members=None, **libcloud_kwargs
):
    """
    Ensures a load balancer is present.

    :param name: Load Balancer name
    :type  name: ``str``

    :param port: Port the load balancer should listen on, defaults to 80
    :type  port: ``str``

    :param protocol: Loadbalancer protocol, defaults to http.
    :type  protocol: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param algorithm: Load balancing algorithm, defaults to ROUND_ROBIN. See Algorithm type
        in Libcloud documentation for a full listing.
    :type algorithm: ``str``

    :param members: An optional list of members to create on deployment
    :type  members: ``list`` of ``dict`` (ip, port)
    """
    balancers = __salt__["libcloud_loadbalancer.list_balancers"](profile)
    match = [z for z in balancers if z["name"] == name]
    if len(match) > 0:
        return state_result(True, "Balancer already exists", name)
    else:
        starting_members = None
        if members is not None:
            starting_members = []
            for m in members:
                starting_members.append({"ip": m["ip"], "port": m["port"]})
        balancer = __salt__["libcloud_loadbalancer.create_balancer"](
            name,
            port,
            protocol,
            profile,
            algorithm=algorithm,
            members=starting_members,
            **libcloud_kwargs
        )
        return state_result(True, "Created new load balancer", name, balancer)


def balancer_absent(name, profile, **libcloud_kwargs):
    """
    Ensures a load balancer is absent.

    :param name: Load Balancer name
    :type  name: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    balancers = __salt__["libcloud_loadbalancer.list_balancers"](profile)
    match = [z for z in balancers if z["name"] == name]
    if len(match) == 0:
        return state_result(True, "Balancer already absent", name)
    else:
        result = __salt__["libcloud_loadbalancer.destroy_balancer"](
            match[0]["id"], profile, **libcloud_kwargs
        )
        return state_result(result, "Deleted load balancer", name)


def member_present(ip, port, balancer_id, profile, **libcloud_kwargs):
    """
    Ensure a load balancer member is present

    :param ip: IP address for the new member
    :type  ip: ``str``

    :param port: Port for the new member
    :type  port: ``int``

    :param balancer_id: id of a load balancer you want to attach the member to
    :type  balancer_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    existing_members = __salt__["libcloud_loadbalancer.list_balancer_members"](
        balancer_id, profile
    )
    for member in existing_members:
        if member["ip"] == ip and member["port"] == port:
            return state_result(True, "Member already present", balancer_id)
    member = __salt__["libcloud_loadbalancer.balancer_attach_member"](
        balancer_id, ip, port, profile, **libcloud_kwargs
    )
    return state_result(
        True,
        "Member added to balancer, id: {}".format(member["id"]),
        balancer_id,
        member,
    )


def member_absent(ip, port, balancer_id, profile, **libcloud_kwargs):
    """
    Ensure a load balancer member is absent, based on IP and Port

    :param ip: IP address for the member
    :type  ip: ``str``

    :param port: Port for the member
    :type  port: ``int``

    :param balancer_id: id of a load balancer you want to detach the member from
    :type  balancer_id: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    existing_members = __salt__["libcloud_loadbalancer.list_balancer_members"](
        balancer_id, profile
    )
    for member in existing_members:
        if member["ip"] == ip and member["port"] == port:
            result = __salt__["libcloud_loadbalancer.balancer_detach_member"](
                balancer_id, member["id"], profile, **libcloud_kwargs
            )
            return state_result(result, "Member removed", balancer_id)
    return state_result(True, "Member already absent", balancer_id)



r"""
Saltclass Configuration
=======================

.. code-block:: yaml

    master_tops:
      saltclass:
        path: /srv/saltclass

Description
===========

This module clones the behaviour of reclass (http://reclass.pantsfullofunix.net/),
without the need of an external app, and add several features to improve flexibility.
Saltclass lets you define your nodes from simple ``yaml`` files (``.yml``) through
hierarchical class inheritance with the possibility to override pillars down the tree.

Features
========

- Define your nodes through hierarchical class inheritance
- Reuse your reclass datas with minimal modifications
    - applications => states
    - parameters => pillars
- Use Jinja templating in your yaml definitions
- Access to the following Salt objects in Jinja
    - ``__opts__``
    - ``__salt__``
    - ``__grains__``
    - ``__pillars__``
    - ``minion_id``
- Chose how to merge or override your lists using ^ character (see examples)
- Expand variables ${} with possibility to escape them if needed \${} (see examples)
- Ignores missing node/class and will simply return empty without breaking the pillar module completely - will be logged

An example subset of datas is available here: http://git.mauras.ch/salt/saltclass/src/master/examples

==========================  ===========
Terms usable in yaml files  Description
==========================  ===========
classes                     A list of classes that will be processed in order
states                      A list of states that will be returned by master_tops function
pillars                     A yaml dictionary that will be returned by the ext_pillar function
environment                 Node saltenv that will be used by master_tops
==========================  ===========

A class consists of:

- zero or more parent classes
- zero or more states
- any number of pillars

A child class can override pillars from a parent class.
A node definition is a class in itself with an added ``environment`` parameter for ``saltenv`` definition.

Class names
===========

Class names mimic salt way of defining states and pillar files.
This means that ``default.users`` class name will correspond to one of these:

- ``<saltclass_path>/classes/default/users.yml``
- ``<saltclass_path>/classes/default/users/init.yml``

Saltclass file hierarchy
========================

A saltclass tree would look like this:

.. code-block:: text

    <saltclass_path>
    ├── classes
    │   ├── app
    │   │   ├── borgbackup.yml
    │   │   └── ssh
    │   │       └── server.yml
    │   ├── default
    │   │   ├── init.yml
    │   │   ├── motd.yml
    │   │   └── users.yml
    │   ├── roles
    │   │   ├── app.yml
    │   │   └── nginx
    │   │       ├── init.yml
    │   │       └── server.yml
    │   └── subsidiaries
    │       ├── gnv.yml
    │       ├── qls.yml
    │       └── zrh.yml
    └── nodes
        ├── geneva
        │   └── gnv.node1.yml
        ├── lausanne
        │   ├── qls.node1.yml
        │   └── qls.node2.yml
        ├── node127.yml
        └── zurich
            ├── zrh.node1.yml
            ├── zrh.node2.yml
            └── zrh.node3.yml


Saltclass Examples
==================

``<saltclass_path>/nodes/lausanne/qls.node1.yml``

.. code-block:: jinja

    environment: base

    classes:
    {% for class in ['default'] %}
      - {{ class }}
    {% endfor %}
      - subsidiaries.{{ __grains__['id'].split('.')[0] }}

``<saltclass_path>/classes/default/init.yml``

.. code-block:: yaml

    classes:
      - default.users
      - default.motd

    states:
      - openssh

    pillars:
      default:
        network:
          dns:
            srv1: 192.168.0.1
            srv2: 192.168.0.2
            domain: example.com
        ntp:
          srv1: 192.168.10.10
          srv2: 192.168.10.20

``<saltclass_path>/classes/subsidiaries/gnv.yml``

.. code-block:: yaml

    pillars:
      default:
        network:
          sub: Geneva
          dns:
            srv1: 10.20.0.1
            srv2: 10.20.0.2
            srv3: 192.168.1.1
            domain: gnv.example.com
        users:
          adm1:
            uid: 1210
            gid: 1210
            gecos: 'Super user admin1'
            homedir: /srv/app/adm1
          adm3:
            uid: 1203
            gid: 1203
            gecos: 'Super user adm

Variable expansions
===================

Escaped variables are rendered as is: ``${test}``

Missing variables are rendered as is: ``${net:dns:srv2}``

.. code-block:: yaml

    pillars:
      app:
      config:
        dns:
          srv1: ${default:network:dns:srv1}
          srv2: ${net:dns:srv2}
        uri: https://application.domain/call?\${test}
        prod_parameters:
          - p1
          - p2
          - p3
      pkg:
        - app-core
        - app-backend

List override
=============

Not using ``^`` as the first entry will simply merge the lists

.. code-block:: yaml

    pillars:
      app:
        pkg:
          - ^
          - app-frontend


.. note:: **Known limitation**

    Currently you can't have both a variable and an escaped variable in the same string as the
    escaped one will not be correctly rendered - '\${xx}' will stay as is instead of being rendered as '${xx}'
"""

import logging

import salt.utils.saltclass as sc

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only run if properly configured
    """
    if __opts__["master_tops"].get("saltclass"):
        return True
    return False


def top(**kwargs):
    """
    Compile tops
    """
    # Node definitions path will be retrieved from args (or set to default),
    # then added to 'salt_data' dict that is passed to the 'get_pillars'
    # function. The dictionary contains:
    #     - __opts__
    #     - __salt__
    #     - __grains__
    #     - __pillar__
    #     - minion_id
    #     - path
    #
    # If successful, the function will return a pillar dict for minion_id.

    # If path has not been set, make a default
    _opts = __opts__["master_tops"]["saltclass"]
    if "path" not in _opts:
        path = "/srv/saltclass"
        log.warning("path variable unset, using default: %s", path)
    else:
        path = _opts["path"]

    # Create a dict that will contain our salt objects
    # to send to get_tops function
    if "id" not in kwargs["opts"]:
        log.warning("Minion id not found - Returning empty dict")
        return {}
    else:
        minion_id = kwargs["opts"]["id"]

    salt_data = {
        "__opts__": kwargs["opts"],
        "__salt__": {},
        "__grains__": kwargs["grains"],
        "__pillar__": {},
        "minion_id": minion_id,
        "path": path,
    }

    return sc.get_tops(minion_id, salt_data)



# Author: Ovidiu Predescu
# Date: July 2011
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Bridges between the Twisted reactor and Tornado IOLoop.

This module lets you run applications and libraries written for
Twisted in a Tornado application.  It can be used in two modes,
depending on which library's underlying event loop you want to use.

This module has been tested with Twisted versions 11.0.0 and newer.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import datetime
import functools
import numbers
import socket
import sys

import twisted.internet.abstract  # type: ignore
from twisted.internet.defer import Deferred  # type: ignore
from twisted.internet.posixbase import PosixReactorBase  # type: ignore
from twisted.internet.interfaces import IReactorFDSet, IDelayedCall, IReactorTime, IReadDescriptor, IWriteDescriptor  # type: ignore
from twisted.python import failure, log  # type: ignore
from twisted.internet import error  # type: ignore
import twisted.names.cache  # type: ignore
import twisted.names.client  # type: ignore
import twisted.names.hosts  # type: ignore
import twisted.names.resolve  # type: ignore

from zope.interface import implementer  # type: ignore

from salt.ext.tornado.concurrent import Future
from salt.ext.tornado.escape import utf8
from salt.ext.tornado import gen
import salt.ext.tornado.ioloop
from salt.ext.tornado.log import app_log
from salt.ext.tornado.netutil import Resolver
from salt.ext.tornado.stack_context import NullContext, wrap
from salt.ext.tornado.ioloop import IOLoop
from salt.ext.tornado.util import timedelta_to_seconds


@implementer(IDelayedCall)
class TornadoDelayedCall(object):
    """DelayedCall object for Tornado."""
    def __init__(self, reactor, seconds, f, *args, **kw):
        self._reactor = reactor
        self._func = functools.partial(f, *args, **kw)
        self._time = self._reactor.seconds() + seconds
        self._timeout = self._reactor._io_loop.add_timeout(self._time,
                                                           self._called)
        self._active = True

    def _called(self):
        self._active = False
        self._reactor._removeDelayedCall(self)
        try:
            self._func()
        except:
            app_log.error("_called caught exception", exc_info=True)

    def getTime(self):
        return self._time

    def cancel(self):
        self._active = False
        self._reactor._io_loop.remove_timeout(self._timeout)
        self._reactor._removeDelayedCall(self)

    def delay(self, seconds):
        self._reactor._io_loop.remove_timeout(self._timeout)
        self._time += seconds
        self._timeout = self._reactor._io_loop.add_timeout(self._time,
                                                           self._called)

    def reset(self, seconds):
        self._reactor._io_loop.remove_timeout(self._timeout)
        self._time = self._reactor.seconds() + seconds
        self._timeout = self._reactor._io_loop.add_timeout(self._time,
                                                           self._called)

    def active(self):
        return self._active


@implementer(IReactorTime, IReactorFDSet)
class TornadoReactor(PosixReactorBase):
    """Twisted reactor built on the Tornado IOLoop.

    `TornadoReactor` implements the Twisted reactor interface on top of
    the Tornado IOLoop.  To use it, simply call `install` at the beginning
    of the application::

        import tornado.platform.twisted
        tornado.platform.twisted.install()
        from twisted.internet import reactor

    When the app is ready to start, call ``IOLoop.current().start()``
    instead of ``reactor.run()``.

    It is also possible to create a non-global reactor by calling
    ``tornado.platform.twisted.TornadoReactor(io_loop)``.  However, if
    the `.IOLoop` and reactor are to be short-lived (such as those used in
    unit tests), additional cleanup may be required.  Specifically, it is
    recommended to call::

        reactor.fireSystemEvent('shutdown')
        reactor.disconnectAll()

    before closing the `.IOLoop`.

    .. versionchanged:: 4.1
       The ``io_loop`` argument is deprecated.
    """
    def __init__(self, io_loop=None):
        if not io_loop:
            io_loop = salt.ext.tornado.ioloop.IOLoop.current()
        self._io_loop = io_loop
        self._readers = {}  # map of reader objects to fd
        self._writers = {}  # map of writer objects to fd
        self._fds = {}  # a map of fd to a (reader, writer) tuple
        self._delayedCalls = {}
        PosixReactorBase.__init__(self)
        self.addSystemEventTrigger('during', 'shutdown', self.crash)

        # IOLoop.start() bypasses some of the reactor initialization.
        # Fire off the necessary events if they weren't already triggered
        # by reactor.run().
        def start_if_necessary():
            if not self._started:
                self.fireSystemEvent('startup')
        self._io_loop.add_callback(start_if_necessary)

    # IReactorTime
    def seconds(self):
        return self._io_loop.time()

    def callLater(self, seconds, f, *args, **kw):
        dc = TornadoDelayedCall(self, seconds, f, *args, **kw)
        self._delayedCalls[dc] = True
        return dc

    def getDelayedCalls(self):
        return [x for x in self._delayedCalls if x._active]

    def _removeDelayedCall(self, dc):
        if dc in self._delayedCalls:
            del self._delayedCalls[dc]

    # IReactorThreads
    def callFromThread(self, f, *args, **kw):
        assert callable(f), "%s is not callable" % f
        with NullContext():
            # This NullContext is mainly for an edge case when running
            # TwistedIOLoop on top of a TornadoReactor.
            # TwistedIOLoop.add_callback uses reactor.callFromThread and
            # should not pick up additional StackContexts along the way.
            self._io_loop.add_callback(f, *args, **kw)

    # We don't need the waker code from the super class, Tornado uses
    # its own waker.
    def installWaker(self):
        pass

    def wakeUp(self):
        pass

    # IReactorFDSet
    def _invoke_callback(self, fd, events):
        if fd not in self._fds:
            return
        (reader, writer) = self._fds[fd]
        if reader:
            err = None
            if reader.fileno() == -1:
                err = error.ConnectionLost()
            elif events & IOLoop.READ:
                err = log.callWithLogger(reader, reader.doRead)
            if err is None and events & IOLoop.ERROR:
                err = error.ConnectionLost()
            if err is not None:
                self.removeReader(reader)
                reader.readConnectionLost(failure.Failure(err))
        if writer:
            err = None
            if writer.fileno() == -1:
                err = error.ConnectionLost()
            elif events & IOLoop.WRITE:
                err = log.callWithLogger(writer, writer.doWrite)
            if err is None and events & IOLoop.ERROR:
                err = error.ConnectionLost()
            if err is not None:
                self.removeWriter(writer)
                writer.writeConnectionLost(failure.Failure(err))

    def addReader(self, reader):
        if reader in self._readers:
            # Don't add the reader if it's already there
            return
        fd = reader.fileno()
        self._readers[reader] = fd
        if fd in self._fds:
            (_, writer) = self._fds[fd]
            self._fds[fd] = (reader, writer)
            if writer:
                # We already registered this fd for write events,
                # update it for read events as well.
                self._io_loop.update_handler(fd, IOLoop.READ | IOLoop.WRITE)
        else:
            with NullContext():
                self._fds[fd] = (reader, None)
                self._io_loop.add_handler(fd, self._invoke_callback,
                                          IOLoop.READ)

    def addWriter(self, writer):
        if writer in self._writers:
            return
        fd = writer.fileno()
        self._writers[writer] = fd
        if fd in self._fds:
            (reader, _) = self._fds[fd]
            self._fds[fd] = (reader, writer)
            if reader:
                # We already registered this fd for read events,
                # update it for write events as well.
                self._io_loop.update_handler(fd, IOLoop.READ | IOLoop.WRITE)
        else:
            with NullContext():
                self._fds[fd] = (None, writer)
                self._io_loop.add_handler(fd, self._invoke_callback,
                                          IOLoop.WRITE)

    def removeReader(self, reader):
        if reader in self._readers:
            fd = self._readers.pop(reader)
            (_, writer) = self._fds[fd]
            if writer:
                # We have a writer so we need to update the IOLoop for
                # write events only.
                self._fds[fd] = (None, writer)
                self._io_loop.update_handler(fd, IOLoop.WRITE)
            else:
                # Since we have no writer registered, we remove the
                # entry from _fds and unregister the handler from the
                # IOLoop
                del self._fds[fd]
                self._io_loop.remove_handler(fd)

    def removeWriter(self, writer):
        if writer in self._writers:
            fd = self._writers.pop(writer)
            (reader, _) = self._fds[fd]
            if reader:
                # We have a reader so we need to update the IOLoop for
                # read events only.
                self._fds[fd] = (reader, None)
                self._io_loop.update_handler(fd, IOLoop.READ)
            else:
                # Since we have no reader registered, we remove the
                # entry from the _fds and unregister the handler from
                # the IOLoop.
                del self._fds[fd]
                self._io_loop.remove_handler(fd)

    def removeAll(self):
        return self._removeAll(self._readers, self._writers)

    def getReaders(self):
        return self._readers.keys()

    def getWriters(self):
        return self._writers.keys()

    # The following functions are mainly used in twisted-style test cases;
    # it is expected that most users of the TornadoReactor will call
    # IOLoop.start() instead of Reactor.run().
    def stop(self):
        PosixReactorBase.stop(self)
        fire_shutdown = functools.partial(self.fireSystemEvent, "shutdown")
        self._io_loop.add_callback(fire_shutdown)

    def crash(self):
        PosixReactorBase.crash(self)
        self._io_loop.stop()

    def doIteration(self, delay):
        raise NotImplementedError("doIteration")

    def mainLoop(self):
        # Since this class is intended to be used in applications
        # where the top-level event loop is ``io_loop.start()`` rather
        # than ``reactor.run()``, it is implemented a little
        # differently than other Twisted reactors. We override
        # ``mainLoop`` instead of ``doIteration`` and must implement
        # timed call functionality on top of `.IOLoop.add_timeout`
        # rather than using the implementation in
        # ``PosixReactorBase``.
        self._io_loop.start()


class _TestReactor(TornadoReactor):
    """Subclass of TornadoReactor for use in unittests.

    This can't go in the test.py file because of import-order dependencies
    with the Twisted reactor test builder.
    """
    def __init__(self):
        # always use a new ioloop
        super(_TestReactor, self).__init__(IOLoop())

    def listenTCP(self, port, factory, backlog=50, interface=''):
        # default to localhost to avoid firewall prompts on the mac
        if not interface:
            interface = '127.0.0.1'
        return super(_TestReactor, self).listenTCP(
            port, factory, backlog=backlog, interface=interface)

    def listenUDP(self, port, protocol, interface='', maxPacketSize=8192):
        if not interface:
            interface = '127.0.0.1'
        return super(_TestReactor, self).listenUDP(
            port, protocol, interface=interface, maxPacketSize=maxPacketSize)


def install(io_loop=None):
    """Install this package as the default Twisted reactor.

    ``install()`` must be called very early in the startup process,
    before most other twisted-related imports. Conversely, because it
    initializes the `.IOLoop`, it cannot be called before
    `.fork_processes` or multi-process `~.TCPServer.start`. These
    conflicting requirements make it difficult to use `.TornadoReactor`
    in multi-process mode, and an external process manager such as
    ``supervisord`` is recommended instead.

    .. versionchanged:: 4.1
       The ``io_loop`` argument is deprecated.

    """
    if not io_loop:
        io_loop = salt.ext.tornado.ioloop.IOLoop.current()
    reactor = TornadoReactor(io_loop)
    from twisted.internet.main import installReactor  # type: ignore
    installReactor(reactor)
    return reactor


@implementer(IReadDescriptor, IWriteDescriptor)
class _FD(object):
    def __init__(self, fd, fileobj, handler):
        self.fd = fd
        self.fileobj = fileobj
        self.handler = handler
        self.reading = False
        self.writing = False
        self.lost = False

    def fileno(self):
        return self.fd

    def doRead(self):
        if not self.lost:
            self.handler(self.fileobj, salt.ext.tornado.ioloop.IOLoop.READ)

    def doWrite(self):
        if not self.lost:
            self.handler(self.fileobj, salt.ext.tornado.ioloop.IOLoop.WRITE)

    def connectionLost(self, reason):
        if not self.lost:
            self.handler(self.fileobj, salt.ext.tornado.ioloop.IOLoop.ERROR)
            self.lost = True

    def logPrefix(self):
        return ''


class TwistedIOLoop(salt.ext.tornado.ioloop.IOLoop):
    """IOLoop implementation that runs on Twisted.

    `TwistedIOLoop` implements the Tornado IOLoop interface on top of
    the Twisted reactor. Recommended usage::

        from salt.ext.tornado.platform.twisted import TwistedIOLoop
        from twisted.internet import reactor
        TwistedIOLoop().install()
        # Set up your tornado application as usual using `IOLoop.instance`
        reactor.run()

    Uses the global Twisted reactor by default.  To create multiple
    ``TwistedIOLoops`` in the same process, you must pass a unique reactor
    when constructing each one.

    Not compatible with `tornado.process.Subprocess.set_exit_callback`
    because the ``SIGCHLD`` handlers used by Tornado and Twisted conflict
    with each other.

    See also :meth:`tornado.ioloop.IOLoop.install` for general notes on
    installing alternative IOLoops.
    """
    def initialize(self, reactor=None, **kwargs):
        super(TwistedIOLoop, self).initialize(**kwargs)
        if reactor is None:
            import twisted.internet.reactor  # type: ignore
            reactor = twisted.internet.reactor
        self.reactor = reactor
        self.fds = {}

    def close(self, all_fds=False):
        fds = self.fds
        self.reactor.removeAll()
        for c in self.reactor.getDelayedCalls():
            c.cancel()
        if all_fds:
            for fd in fds.values():
                self.close_fd(fd.fileobj)

    def add_handler(self, fd, handler, events):
        if fd in self.fds:
            raise ValueError('fd %s added twice' % fd)
        fd, fileobj = self.split_fd(fd)
        self.fds[fd] = _FD(fd, fileobj, wrap(handler))
        if events & salt.ext.tornado.ioloop.IOLoop.READ:
            self.fds[fd].reading = True
            self.reactor.addReader(self.fds[fd])
        if events & salt.ext.tornado.ioloop.IOLoop.WRITE:
            self.fds[fd].writing = True
            self.reactor.addWriter(self.fds[fd])

    def update_handler(self, fd, events):
        fd, fileobj = self.split_fd(fd)
        if events & salt.ext.tornado.ioloop.IOLoop.READ:
            if not self.fds[fd].reading:
                self.fds[fd].reading = True
                self.reactor.addReader(self.fds[fd])
        else:
            if self.fds[fd].reading:
                self.fds[fd].reading = False
                self.reactor.removeReader(self.fds[fd])
        if events & salt.ext.tornado.ioloop.IOLoop.WRITE:
            if not self.fds[fd].writing:
                self.fds[fd].writing = True
                self.reactor.addWriter(self.fds[fd])
        else:
            if self.fds[fd].writing:
                self.fds[fd].writing = False
                self.reactor.removeWriter(self.fds[fd])

    def remove_handler(self, fd):
        fd, fileobj = self.split_fd(fd)
        if fd not in self.fds:
            return
        self.fds[fd].lost = True
        if self.fds[fd].reading:
            self.reactor.removeReader(self.fds[fd])
        if self.fds[fd].writing:
            self.reactor.removeWriter(self.fds[fd])
        del self.fds[fd]

    def start(self):
        old_current = IOLoop.current(instance=False)
        try:
            self._setup_logging()
            self.make_current()
            self.reactor.run()
        finally:
            if old_current is None:
                IOLoop.clear_current()
            else:
                old_current.make_current()

    def stop(self):
        self.reactor.crash()

    def add_timeout(self, deadline, callback, *args, **kwargs):
        # This method could be simplified (since tornado 4.0) by
        # overriding call_at instead of add_timeout, but we leave it
        # for now as a test of backwards-compatibility.
        if isinstance(deadline, numbers.Real):
            delay = max(deadline - self.time(), 0)
        elif isinstance(deadline, datetime.timedelta):
            delay = timedelta_to_seconds(deadline)
        else:
            raise TypeError("Unsupported deadline %r")
        return self.reactor.callLater(
            delay, self._run_callback,
            functools.partial(wrap(callback), *args, **kwargs))

    def remove_timeout(self, timeout):
        if timeout.active():
            timeout.cancel()

    def add_callback(self, callback, *args, **kwargs):
        self.reactor.callFromThread(
            self._run_callback,
            functools.partial(wrap(callback), *args, **kwargs))

    def add_callback_from_signal(self, callback, *args, **kwargs):
        self.add_callback(callback, *args, **kwargs)


class TwistedResolver(Resolver):
    """Twisted-based asynchronous resolver.

    This is a non-blocking and non-threaded resolver.  It is
    recommended only when threads cannot be used, since it has
    limitations compared to the standard ``getaddrinfo``-based
    `~tornado.netutil.Resolver` and
    `~tornado.netutil.ThreadedResolver`.  Specifically, it returns at
    most one result, and arguments other than ``host`` and ``family``
    are ignored.  It may fail to resolve when ``family`` is not
    ``socket.AF_UNSPEC``.

    Requires Twisted 12.1 or newer.

    .. versionchanged:: 4.1
       The ``io_loop`` argument is deprecated.
    """
    def initialize(self, io_loop=None):
        self.io_loop = io_loop or IOLoop.current()
        # partial copy of twisted.names.client.createResolver, which doesn't
        # allow for a reactor to be passed in.
        self.reactor = salt.ext.tornado.platform.twisted.TornadoReactor(io_loop)

        host_resolver = twisted.names.hosts.Resolver('/etc/hosts')
        cache_resolver = twisted.names.cache.CacheResolver(reactor=self.reactor)
        real_resolver = twisted.names.client.Resolver('/etc/resolv.conf',
                                                      reactor=self.reactor)
        self.resolver = twisted.names.resolve.ResolverChain(
            [host_resolver, cache_resolver, real_resolver])

    @gen.coroutine
    def resolve(self, host, port, family=0):
        # getHostByName doesn't accept IP addresses, so if the input
        # looks like an IP address just return it immediately.
        if twisted.internet.abstract.isIPAddress(host):
            resolved = host
            resolved_family = socket.AF_INET
        elif twisted.internet.abstract.isIPv6Address(host):
            resolved = host
            resolved_family = socket.AF_INET6
        else:
            deferred = self.resolver.getHostByName(utf8(host))
            resolved = yield gen.Task(deferred.addBoth)
            if isinstance(resolved, failure.Failure):
                try:
                    resolved.raiseException()
                except twisted.names.error.DomainError as e:
                    raise IOError(e)
            elif twisted.internet.abstract.isIPAddress(resolved):
                resolved_family = socket.AF_INET
            elif twisted.internet.abstract.isIPv6Address(resolved):
                resolved_family = socket.AF_INET6
            else:
                resolved_family = socket.AF_UNSPEC
        if family != socket.AF_UNSPEC and family != resolved_family:
            raise Exception('Requested socket family %d but got %d' %
                            (family, resolved_family))
        result = [
            (resolved_family, (resolved, port)),
        ]
        raise gen.Return(result)


if hasattr(gen.convert_yielded, 'register'):
    @gen.convert_yielded.register(Deferred)  # type: ignore
    def _(d):
        f = Future()

        def errback(failure):
            try:
                failure.raiseException()
                # Should never happen, but just in case
                raise Exception("errback called without error")
            except:
                f.set_exc_info(sys.exc_info())
        d.addCallbacks(f.set_result, errback)
        return f



import os
import time

import pytest

from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS


@pytest.mark.windows_whitelisted
class LoaderOverridesTest(ModuleCase):
    def setUp(self):
        self.run_function("saltutil.sync_modules")

    @pytest.mark.slow_test
    @pytest.mark.timeout_unless_on_windows(120)
    def test_overridden_internal(self):
        # To avoid a race condition on Windows, we need to make sure the
        # `override_test.py` file is present in the _modules directory before
        # trying to list all functions. This test may execute before the
        # minion has finished syncing down the files it needs.
        module = os.path.join(
            RUNTIME_VARS.TMP,
            "rootdir",
            "cache",
            "files",
            "base",
            "_modules",
            "override_test.py",
        )
        tries = 0
        while not os.path.exists(module):
            tries += 1
            if tries > 60:
                break
            time.sleep(1)

        funcs = self.run_function("sys.list_functions")

        # We placed a test module under _modules.
        # The previous functions should also still exist.
        self.assertIn("test.ping", funcs)

        # A non existing function should, of course, not exist
        self.assertNotIn("brain.left_hemisphere", funcs)

        # There should be a new function for the test module, recho
        self.assertIn("test.recho", funcs)

        text = "foo bar baz quo qux"
        self.assertEqual(
            self.run_function("test.echo", arg=[text])[::-1],
            self.run_function("test.recho", arg=[text]),
        )



"""
Manage the Thorium complex event reaction system
"""

import salt.thorium


def start(grains=False, grain_keys=None, pillar=False, pillar_keys=None):
    """
    Execute the Thorium runtime
    """
    state = salt.thorium.ThorState(__opts__, grains, grain_keys, pillar, pillar_keys)
    state.start_runtime()



n9k_show_ver = """
Cisco Nexus Operating System (NX-OS) Software
TAC support: http://www.cisco.com/tac
Copyright (C) 2002-2018, Cisco and/or its affiliates.
All rights reserved.
The copyrights to certain works contained in this software are
owned by other third parties and used and distributed under their own
licenses, such as open source.  This software is provided "as is," and unless
otherwise stated, there is no warranty, express or implied, including but not
limited to warranties of merchantability and fitness for a particular purpose.
Certain components of this software are licensed under
the GNU General Public License (GPL) version 2.0 or
GNU General Public License (GPL) version 3.0  or the GNU
Lesser General Public License (LGPL) Version 2.1 or
Lesser General Public License (LGPL) Version 2.0.
A copy of each such license is available at
http://www.opensource.org/licenses/gpl-2.0.php and
http://opensource.org/licenses/gpl-3.0.html and
http://www.opensource.org/licenses/lgpl-2.1.php and
http://www.gnu.org/licenses/old-licenses/library.txt.

Software
  BIOS: version 08.36
  NXOS: version 9.2(1)
  BIOS compile time:  06/07/2019
  NXOS image file is: bootflash:///nxos.9.2.1.bin
  NXOS compile time:  7/17/2018 16:00:00 [07/18/2018 00:21:19]


Hardware
  cisco Nexus9000 C9504 (4 Slot) Chassis ("Supervisor Module")
  Intel(R) Xeon(R) CPU E5-2403 0 @ 1.80GHz with 16400084 kB of memory.
  Processor Board ID SAL1909A7VC

  Device name: n9k-device
  bootflash:   53298520 kB
Kernel uptime is 0 day(s), 18 hour(s), 26 minute(s), 18 second(s)

Last reset at 931765 usecs after Mon Dec  2 01:21:36 2019
  Reason: Reset Requested by CLI command reload
  System version: 9.2(4)
  Service:

plugin
  Core Plugin, Ethernet Plugin

Active Package(s):
"""

n9k_show_ver_structured = [
    {
        "header_str": (
            "Cisco Nexus Operating System (NX-OS) Software\nTAC support:"
            " http://www.cisco.com/tac\nCopyright (C) 2002-2019, Cisco and/or its"
            " affiliates.\nAll rights reserved.\nThe copyrights to certain works"
            " contained in this software are\nowned by other third parties and used and"
            " distributed under their own\nlicenses, such as open source.  This"
            ' software is provided "as is," and unless\notherwise stated, there is no'
            " warranty, express or implied, including but not\nlimited to warranties of"
            " merchantability and fitness for a particular purpose.\nCertain components"
            " of this software are licensed under\nthe GNU General Public License (GPL)"
            " version 2.0 or \nGNU General Public License (GPL) version 3.0  or the"
            " GNU\nLesser General Public License (LGPL) Version 2.1 or \nLesser General"
            " Public License (LGPL) Version 2.0. \nA copy of each such license is"
            " available at\nhttp://www.opensource.org/licenses/gpl-2.0.php"
            " and\nhttp://opensource.org/licenses/gpl-3.0.html"
            " and\nhttp://www.opensource.org/licenses/lgpl-2.1.php"
            " and\nhttp://www.gnu.org/licenses/old-licenses/library.txt.\n"
        ),
        "bios_ver_str": "07.66",
        "kickstart_ver_str": "7.0(3)I7(8) [build 7.0(3)I7(7.16)]",
        "bios_cmpl_time": "06/12/2019",
        "kick_file_name": "bootflash:///nxos.7.0.3.I7.7.16.bin",
        "kick_cmpl_time": " 11/29/2019 13:00:00",
        "kick_tmstmp": "11/29/2019 21:52:12",
        "chassis_id": "Nexus9000 C9396PX Chassis",
        "cpu_name": "Intel(R) Core(TM) i3- CPU @ 2.50GHz",
        "memory": 16401088,
        "mem_type": "kB",
        "proc_board_id": "SAL1821T9EF",
        "host_name": "dt-n9k5-1",
        "bootflash_size": 21693714,
        "kern_uptm_days": 1,
        "kern_uptm_hrs": 23,
        "kern_uptm_mins": 32,
        "kern_uptm_secs": 47,
        "rr_usecs": 915186,
        "rr_ctime": "Sun Dec  1 15:47:17 2019",
        "rr_reason": "Reset Requested by CLI command reload",
        "rr_sys_ver": "9.2(4)",
        "rr_service": "",
        "manufacturer": "Cisco Systems, Inc.",
        "TABLE_package_list": {"ROW_package_list": {"package_id": {}}},
    }
]

# Data returned from nxapi for raw text is in list form.
n9k_show_ver_list = [
    "Cisco Nexus Operating System (NX-OS) Software\nTAC support:"
    " http://www.cisco.com/tac\nCopyright (C) 2002-2019, Cisco and/or its"
    " affiliates.\nAll rights reserved.\nThe copyrights to certain works contained in"
    " this software are\nowned by other third parties and used and distributed under"
    ' their own\nlicenses, such as open source.  This software is provided "as is," and'
    " unless\notherwise stated, there is no warranty, express or implied, including but"
    " not\nlimited to warranties of merchantability and fitness for a particular"
    " purpose.\nCertain components of this software are licensed under\nthe GNU General"
    " Public License (GPL) version 2.0 or \nGNU General Public License (GPL) version"
    " 3.0  or the GNU\nLesser General Public License (LGPL) Version 2.1 or \nLesser"
    " General Public License (LGPL) Version 2.0. \nA copy of each such license is"
    " available at\nhttp://www.opensource.org/licenses/gpl-2.0.php"
    " and\nhttp://opensource.org/licenses/gpl-3.0.html"
    " and\nhttp://www.opensource.org/licenses/lgpl-2.1.php"
    " and\nhttp://www.gnu.org/licenses/old-licenses/library.txt.\n\nSoftware\n  BIOS:"
    " version 07.66\n  NXOS: version 7.0(3)I7(8) [build 7.0(3)I7(7.16)]\n  BIOS compile"
    " time:  06/12/2019\n  NXOS image file is: bootflash:///nxos.7.0.3.I7.7.16.bin\n "
    " NXOS compile time:  11/29/2019 13:00:00 [11/29/2019 21:52:12]\n\n\nHardware\n "
    " cisco Nexus9000 C9396PX Chassis \n  Intel(R) Core(TM) i3- CPU @ 2.50GHz with"
    " 16401088 kB of memory.\n  Processor Board ID SAL1821T9EF\n\n  Device name:"
    " n9k-device\n  bootflash:   21693714 kB\nKernel uptime is 1 day(s), 22 hour(s), 54"
    " minute(s), 13 second(s)\n\nLast reset at 915186 usecs after Sun Dec  1 15:47:17"
    " 2019\n  Reason: Reset Requested by CLI command reload\n  System version: 9.2(4)\n"
    "  Service: \n\nplugin\n  Core Plugin, Ethernet Plugin\n\nActive Package(s):\n \n"
]

n9k_show_user_account = """
user:salt_test
        this user account has no expiry date
        roles:network-operator network-admin dev-ops
"""

n9k_show_user_account_list = [
    "user:salt_test\n        this user account has no expiry date\n       "
    " roles:network-operator dev-ops network-admin \n"
]

n9k_show_ver_int_list = [
    "Cisco Nexus Operating System (NX-OS) Software\nTAC support:"
    " http://www.cisco.com/tac\nCopyright (C) 2002-2019, Cisco and/or its"
    " affiliates.\nAll rights reserved.\nThe copyrights to certain works contained in"
    " this software are\nowned by other third parties and used and distributed under"
    ' their own\nlicenses, such as open source.  This software is provided "as is," and'
    " unless\notherwise stated, there is no warranty, express or implied, including but"
    " not\nlimited to warranties of merchantability and fitness for a particular"
    " purpose.\nCertain components of this software are licensed under\nthe GNU General"
    " Public License (GPL) version 2.0 or \nGNU General Public License (GPL) version"
    " 3.0  or the GNU\nLesser General Public License (LGPL) Version 2.1 or \nLesser"
    " General Public License (LGPL) Version 2.0. \nA copy of each such license is"
    " available at\nhttp://www.opensource.org/licenses/gpl-2.0.php"
    " and\nhttp://opensource.org/licenses/gpl-3.0.html"
    " and\nhttp://www.opensource.org/licenses/lgpl-2.1.php"
    " and\nhttp://www.gnu.org/licenses/old-licenses/library.txt.\n\nSoftware\n  BIOS:"
    " version 07.66\n  NXOS: version 7.0(3)I7(8) [build 7.0(3)I7(7.16)]\n  BIOS compile"
    " time:  06/12/2019\n  NXOS image file is: bootflash:///nxos.7.0.3.I7.7.16.bin\n "
    " NXOS compile time:  11/29/2019 13:00:00 [11/29/2019 21:52:12]\n\n\nHardware\n "
    " cisco Nexus9000 C9396PX Chassis \n  Intel(R) Core(TM) i3- CPU @ 2.50GHz with"
    " 16401088 kB of memory.\n  Processor Board ID SAL1821T9EF\n\n  Device name:"
    " n9k-device\n  bootflash:   21693714 kB\nKernel uptime is 1 day(s), 23 hour(s), 24"
    " minute(s), 35 second(s)\n\nLast reset at 915186 usecs after Sun Dec  1 15:47:17"
    " 2019\n  Reason: Reset Requested by CLI command reload\n  System version: 9.2(4)\n"
    "  Service: \n\nplugin\n  Core Plugin, Ethernet Plugin\n\nActive Package(s):\n \n",
    "Ethernet1/1 is down (XCVR not inserted)\nadmin state is down, Dedicated"
    " Interface\n  Hardware: 1000/10000 Ethernet, address: 88f0.31dc.7de6 (bia"
    " 88f0.31dc.7de6)\n  MTU 1500 bytes, BW 10000000 Kbit, DLY 10 usec\n  reliability"
    " 255/255, txload 1/255, rxload 1/255\n  Encapsulation ARPA, medium is broadcast\n "
    " Port mode is trunk\n  auto-duplex, auto-speed\n  Beacon is turned off\n "
    " Auto-Negotiation is turned on  FEC mode is Auto\n  Input flow-control is off,"
    " output flow-control is off\n  Auto-mdix is turned off\n  Switchport monitor is"
    " off \n  EtherType is 0x8100 \n  EEE (efficient-ethernet) : n/a\n    admin fec"
    " state is auto, oper fec state is off\n  Last link flapped never\n  Last clearing"
    ' of "show interface" counters 1d21h\n  0 interface resets\n  Load-Interval #1: 30'
    " seconds\n    30 seconds input rate 0 bits/sec, 0 packets/sec\n    30 seconds"
    " output rate 0 bits/sec, 0 packets/sec\n    input rate 0 bps, 0 pps; output rate 0"
    " bps, 0 pps\n  Load-Interval #2: 5 minute (300 seconds)\n    300 seconds input"
    " rate 0 bits/sec, 0 packets/sec\n    300 seconds output rate 0 bits/sec, 0"
    " packets/sec\n    input rate 0 bps, 0 pps; output rate 0 bps, 0 pps\n  RX\n    0"
    " unicast packets  0 multicast packets  0 broadcast packets\n    0 input packets  0"
    " bytes\n    0 jumbo packets  0 storm suppression packets\n    0 runts  0 giants  0"
    " CRC  0 no buffer\n    0 input error  0 short frame  0 overrun   0 underrun  0"
    " ignored\n    0 watchdog  0 bad etype drop  0 bad proto drop  0 if down drop\n   "
    " 0 input with dribble  0 input discard\n    0 Rx pause\n  TX\n    0 unicast"
    " packets  0 multicast packets  0 broadcast packets\n    0 output packets  0"
    " bytes\n    0 jumbo packets\n    0 output error  0 collision  0 deferred  0 late"
    " collision\n    0 lost carrier  0 no carrier  0 babble  0 output discard\n    0 Tx"
    " pause\n\n",
]

n9k_show_ver_int_list_structured = [
    {
        "header_str": (
            "Cisco Nexus Operating System (NX-OS) Software\nTAC support:"
            " http://www.cisco.com/tac\nCopyright (C) 2002-2019, Cisco and/or its"
            " affiliates.\nAll rights reserved.\nThe copyrights to certain works"
            " contained in this software are\nowned by other third parties and used and"
            " distributed under their own\nlicenses, such as open source.  This"
            ' software is provided "as is," and unless\notherwise stated, there is no'
            " warranty, express or implied, including but not\nlimited to warranties of"
            " merchantability and fitness for a particular purpose.\nCertain components"
            " of this software are licensed under\nthe GNU General Public License (GPL)"
            " version 2.0 or \nGNU General Public License (GPL) version 3.0  or the"
            " GNU\nLesser General Public License (LGPL) Version 2.1 or \nLesser General"
            " Public License (LGPL) Version 2.0. \nA copy of each such license is"
            " available at\nhttp://www.opensource.org/licenses/gpl-2.0.php"
            " and\nhttp://opensource.org/licenses/gpl-3.0.html"
            " and\nhttp://www.opensource.org/licenses/lgpl-2.1.php"
            " and\nhttp://www.gnu.org/licenses/old-licenses/library.txt.\n"
        ),
        "bios_ver_str": "07.66",
        "kickstart_ver_str": "7.0(3)I7(8) [build 7.0(3)I7(7.16)]",
        "bios_cmpl_time": "06/12/2019",
        "kick_file_name": "bootflash:///nxos.7.0.3.I7.7.16.bin",
        "kick_cmpl_time": " 11/29/2019 13:00:00",
        "kick_tmstmp": "11/29/2019 21:52:12",
        "chassis_id": "Nexus9000 C9396PX Chassis",
        "cpu_name": "Intel(R) Core(TM) i3- CPU @ 2.50GHz",
        "memory": 16401088,
        "mem_type": "kB",
        "proc_board_id": "SAL1821T9EF",
        "host_name": "n9k-device",
        "bootflash_size": 21693714,
        "kern_uptm_days": 1,
        "kern_uptm_hrs": 23,
        "kern_uptm_mins": 31,
        "kern_uptm_secs": 18,
        "rr_usecs": 915186,
        "rr_ctime": "Sun Dec  1 15:47:17 2019",
        "rr_reason": "Reset Requested by CLI command reload",
        "rr_sys_ver": "9.2(4)",
        "rr_service": "",
        "manufacturer": "Cisco Systems, Inc.",
        "TABLE_package_list": {"ROW_package_list": {"package_id": {}}},
    },
    {
        "TABLE_interface": {
            "ROW_interface": {
                "interface": "Ethernet1/1",
                "state": "down",
                "state_rsn_desc": "XCVR not inserted",
                "admin_state": "down",
                "share_state": "Dedicated",
                "eth_hw_desc": "1000/10000 Ethernet",
                "eth_hw_addr": "88f0.31dc.7de6",
                "eth_bia_addr": "88f0.31dc.7de6",
                "eth_mtu": "1500",
                "eth_bw": 10000000,
                "eth_dly": 10,
                "eth_reliability": "255",
                "eth_txload": "1",
                "eth_rxload": "1",
                "medium": "broadcast",
                "eth_mode": "trunk",
                "eth_duplex": "auto",
                "eth_speed": "auto-speed",
                "eth_beacon": "off",
                "eth_autoneg": "on",
                "eth_in_flowctrl": "off",
                "eth_out_flowctrl": "off",
                "eth_mdix": "off",
                "eth_swt_monitor": "off",
                "eth_ethertype": "0x8100",
                "eth_eee_state": "n/a",
                "eth_admin_fec_state": "auto",
                "eth_oper_fec_state": "off",
                "eth_link_flapped": "never",
                "eth_clear_counters": "1d21h",
                "eth_reset_cntr": 0,
                "eth_load_interval1_rx": 30,
                "eth_inrate1_bits": "0",
                "eth_inrate1_pkts": "0",
                "eth_load_interval1_tx": "30",
                "eth_outrate1_bits": "0",
                "eth_outrate1_pkts": "0",
                "eth_inrate1_summary_bits": "0 bps",
                "eth_inrate1_summary_pkts": "0 pps",
                "eth_outrate1_summary_bits": "0 bps",
                "eth_outrate1_summary_pkts": "0 pps",
                "eth_load_interval2_rx": "300",
                "eth_inrate2_bits": "0",
                "eth_inrate2_pkts": "0",
                "eth_load_interval2_tx": "300",
                "eth_outrate2_bits": "0",
                "eth_outrate2_pkts": "0",
                "eth_inrate2_summary_bits": "0 bps",
                "eth_inrate2_summary_pkts": "0 pps",
                "eth_outrate2_summary_bits": "0 bps",
                "eth_outrate2_summary_pkts": "0 pps",
                "eth_inucast": 0,
                "eth_inmcast": 0,
                "eth_inbcast": 0,
                "eth_inpkts": 0,
                "eth_inbytes": 0,
                "eth_jumbo_inpkts": "0",
                "eth_storm_supp": "0",
                "eth_runts": 0,
                "eth_giants": 0,
                "eth_crc": "0",
                "eth_nobuf": 0,
                "eth_inerr": "0",
                "eth_frame": "0",
                "eth_overrun": "0",
                "eth_underrun": "0",
                "eth_ignored": "0",
                "eth_watchdog": "0",
                "eth_bad_eth": "0",
                "eth_bad_proto": "0",
                "eth_in_ifdown_drops": "0",
                "eth_dribble": "0",
                "eth_indiscard": "0",
                "eth_inpause": "0",
                "eth_outucast": 0,
                "eth_outmcast": 0,
                "eth_outbcast": 0,
                "eth_outpkts": 0,
                "eth_outbytes": 0,
                "eth_jumbo_outpkts": "0",
                "eth_outerr": "0",
                "eth_coll": "0",
                "eth_deferred": "0",
                "eth_latecoll": "0",
                "eth_lostcarrier": "0",
                "eth_nocarrier": "0",
                "eth_babbles": "0",
                "eth_outdiscard": "0",
                "eth_outpause": "0",
            }
        }
    },
]

n9k_get_user_output = (
    "username salt_test password 5"
    " $5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC  role network-operator"
)
